"""
Build and validate a linear model
=================================
"""

# %%
# In this example, we  build a linear regression model and validate it numerically and graphically.
#
# The linear model links a scalar variable :math:`Y` and to an n-dimensional one :math:`\vect{X} = (X_i)_{i \leq n}`, as follows:
#
# .. math::
#    \tilde{Y} = a_0 + \sum_{i=1}^n a_i X_i + \varepsilon
#
# where :math:`\varepsilon` is the residual, supposed to follow :math:`\mathcal{N}(0.0, 1.0)`.
#
# The linear model may be validated graphically if :math:`\vect{X}` is of dimension 1, by drawing on the same graph the cloud :math:`(X_i, Y_i)`.
#
# The linear model can also be validated numerically with several tests:
#
# - `LinearModelFisher`: tests the nullity of the regression linear model coefficients (Fisher distribution used),
# - `LinearModelResidualMean`: tests, under the hypothesis of a Gaussian sample, if the mean of the residual is equal to zero.
#   It is based on the Student test (equality of mean for two Gaussian samples).
#
#
# The hypothesis on the residuals (centered Gaussian distribution) may be validated:
#
# - graphically if :math:`\vect{X}` is of dimension 1, by drawing the residual couples (:math:`\varepsilon_i, \varepsilon_{i+1}`),
#   where the residual :math:`\varepsilon_i` is evaluated on the samples :math:`(X, Y)`.
# - numerically with the `LinearModelResidualMean` test which tests, under the hypothesis of a Gaussian sample, if the mean of the residual is equal to zero.
#   It is based on the Student test (equality of mean for two Gaussian samples).
#

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# Generate `X, Y` samples
N = 1000
Xsample = ot.Triangular(1.0, 5.0, 10.0).getSample(N)
Ysample = Xsample * 3.0 + ot.Normal(0.5, 1.0).getSample(N)

# %%
# Generate a particular scalar sampleX
particularXSample = ot.Triangular(1.0, 5.0, 10.0).getSample(N)

# %%
# Create the linear model from `Y, X` samples
result = ot.LinearModelAlgorithm(Xsample, Ysample).getResult()

# Get the coefficients ai
print("coefficients of the linear regression model = ", result.getCoefficients())

# Get the confidence intervals of the `ai` coefficients
print(
    "confidence intervals of the coefficients = ",
    ot.LinearModelAnalysis(result).getCoefficientsConfidenceInterval(0.9),
)


# %%
# Validate the model with a visual test
graph = ot.VisualTest.DrawLinearModel(Xsample, Ysample, result)
view = otv.View(graph)

# %%
# Draw the graph of the residual values
graph = ot.VisualTest.DrawLinearModelResidual(Xsample, Ysample, result)
view = otv.View(graph)

# %%
# Check the nullity of the regression linear model coefficients
resultLinearModelFisher = ot.LinearModelTest.LinearModelFisher(
    Xsample, Ysample, result, 0.10
)
print("Test Success ? ", resultLinearModelFisher.getBinaryQualityMeasure())
print("p-value of the LinearModelFisher Test = ", resultLinearModelFisher.getPValue())
print("p-value threshold = ", resultLinearModelFisher.getThreshold())

# %%
# Check, under the hypothesis of a Gaussian sample, if the mean of the residuals is equal to zero
resultLinearModelResidualMean = ot.LinearModelTest.LinearModelResidualMean(
    Xsample, Ysample, result, 0.10
)
print("Test Success ? ", resultLinearModelResidualMean.getBinaryQualityMeasure())
print(
    "p-value of the LinearModelResidualMean Test = ",
    resultLinearModelResidualMean.getPValue(),
)
print("p-value threshold = ", resultLinearModelResidualMean.getThreshold())


# %%
otv.View.ShowAll()



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

sample = ot.Normal(2).getSample(30)

# Create an empty graph
graph = ot.Graph("Text example", "x1", "x2")

# Create the cloud
cloud = ot.Cloud(sample, "blue", "fsquare", "")

# Then, draw it
graph.add(cloud)

# Display extrema indices
x1 = [x[0] for x in sample[:, 0]]
x2 = [x[0] for x in sample[:, 1]]
idx = [0] * 4
idx[0] = x1.index(min(x1))
idx[1] = x1.index(max(x1))
idx[2] = x2.index(min(x2))
idx[3] = x2.index(max(x2))

labels = ot.Description(sample.getSize())
for i in range(4):
    labels[idx[i]] = str(idx[i])

position = ot.Description(sample.getSize(), "top")
position[idx[0]] = "right"
position[idx[1]] = "left"
position[idx[2]] = "top"
position[idx[3]] = "bottom"

annotations = ot.Text(sample, labels)
annotations.setColor("red")
annotations.setTextPositions(position)

graph.add(annotations)

fig = plt.figure(figsize=(4, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=False)



"""
Use case : Ishigami test function
=================================
"""

import openturns as ot
import math as m


class IshigamiModel:
    """
    Data class for the Ishigami model.


    Attributes
    ----------


    dim : The dimension of the problem
          dim = 3

    a : float
        Constant: a = 7.0

    b : float
        Constant: b = 0.1

    X1 : :class:`~openturns.Uniform`
        First marginal, ot.Uniform(-np.pi, np.pi)

    X2 : :class:`~openturns.Uniform`
        Second marginal, ot.Uniform(-np.pi, np.pi)

    X3 : :class:`~openturns.Uniform`
        Third marginal, ot.Uniform(-np.pi, np.pi)

    distribution : :class:`~openturns.JointDistribution`
        The joint distribution of the input parameters.

    ishigami : :class:`~openturns.SymbolicFunction`
        The Ishigami model with a, b as variables.

    model : :class:`~openturns.ParametricFunction`
        The Ishigami model with the a=7.0 and b=0.1 parameters fixed.

    expectation : float
        Expectation of the output variable.

    variance : float
        Variance of the output variable.

    S1 : float
         First order Sobol index number 1

    S2 : float
         First order Sobol index number 2

    S3 : float
         First order Sobol index number 3

    S12 : float
          Second order Sobol index for marginals 1 and 2.

    S13 : float
          Second order Sobol index for marginals 1 and 3.

    S23 : float
          Second order Sobol index for marginals 2 and 3.

    S123 : float

    ST1 : float
          Total order Sobol index number 1.

    ST2 : float
          Total order Sobol index number 2.

    ST3 : float
          Total order Sobol index number 3.


    Examples
    --------
    >>> from openturns.usecases import ishigami_function
    >>> # Load the Ishigami model
    >>> im = ishigami_function.IshigamiModel()
    """

    def __init__(self):
        # dimension
        self.dim = 3
        # Fixed parameters for the Ishigami function
        self.a = 7.0
        self.b = 0.1

        # First marginal : X1
        self.X1 = ot.Uniform(-m.pi, m.pi)
        self.X1.setName("X1")
        self.X1.setDescription(["X1"])

        # Second marginal : X2
        self.X2 = ot.Uniform(-m.pi, m.pi)
        self.X2.setName("X2")
        self.X2.setDescription(["X2"])

        # Third marginal : X3
        self.X3 = ot.Uniform(-m.pi, m.pi)
        self.X3.setName("X3")
        self.X3.setDescription(["X3"])

        # Input distribution
        self.distribution = ot.JointDistribution([self.X1, self.X2, self.X3])
        self.distribution.setDescription(["X1", "X2", "X3"])
        self.inputDistribution = self.distribution  # deprecated

        self.ishigami = ot.SymbolicFunction(
            ["X1", "X2", "X3", "a", "b"],
            ["sin(X1) + a * sin(X2)^2 + b * X3^4 * sin(X1)"],
        )
        self.ishigami.setOutputDescription(["y"])
        # The Ishigami model
        self.model = ot.ParametricFunction(self.ishigami, [3, 4], [self.a, self.b])

        self.expectation = self.a / 2.0
        self.variance = (
            1.0 / 2
            + self.a**2 / 8.0
            + self.b * m.pi**4 / 5.0
            + self.b**2 * m.pi**8 / 18.0
        )
        self.S1 = (
            1.0 / 2.0 + self.b * m.pi**4 / 5.0 + self.b**2 * m.pi**8 / 50.0
        ) / self.variance
        self.S2 = (self.a**2 / 8.0) / self.variance
        self.S3 = 0.0
        self.S12 = 0.0
        self.S13 = self.b**2 * m.pi**8 / 2.0 * (1.0 / 9.0 - 1.0 / 25.0) / self.variance
        self.S23 = 0.0
        self.S123 = 0.0
        self.ST1 = self.S1 + self.S13
        self.ST2 = self.S2
        self.ST3 = self.S3 + self.S13



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


def cleanScalar(inScalar):
    if abs(inScalar) < 1.0e-10:
        inScalar = 0.0
    return inScalar


#
# Physical model
#

dim = 26
inputFunction = ot.Description(dim)
inputFunction[0] = "r"
for i in range(1, dim):
    inputFunction[i] = "s" + str(i)

formulas = ot.Description(1)
formulas[0] = "r"
for i in range(1, dim):
    formulas[0] = formulas[0] + " - (" + inputFunction[i] + " ^ 2.0) / " + str(i)
limitState = ot.SymbolicFunction(inputFunction, formulas)

#
# Probabilistic model
#

mean = ot.Point(dim, 0.2)
mean[0] = 0.5

sigma = ot.Point(dim, 0.1)

R = ot.CorrelationMatrix(dim)
myDistribution = ot.Normal(mean, sigma, R)

#
# Limit state
#

vect = ot.RandomVector(myDistribution)

output = ot.CompositeRandomVector(limitState, vect)

myEvent = ot.ThresholdEvent(output, ot.Less(), 0.0)

#
# Calculs
#

#
# FORM/SORM Cobyla
myCobyla = ot.Cobyla()
myCobyla.setMaximumCallsNumber(100 * dim)
myCobyla.setMaximumAbsoluteError(1.0e-10)
myCobyla.setMaximumRelativeError(1.0e-10)
myCobyla.setMaximumResidualError(1.0e-10)
myCobyla.setMaximumConstraintError(1.0e-10)
myCobyla.setStartingPoint(myDistribution.getMean())

myAlgoC = ot.FORM(myCobyla, myEvent)
myAlgoC2 = ot.SORM(myCobyla, myEvent)

myAlgoC.run()
myAlgoC2.run()

resultC = myAlgoC.getResult()
resultC2 = myAlgoC2.getResult()

#
# FORM/SORM Abdo Rackwitz
myAbdoRackwitz = ot.AbdoRackwitz()
myAbdoRackwitz.setMaximumIterationNumber(100 * dim)
myAbdoRackwitz.setMaximumAbsoluteError(1.0e-10)
myAbdoRackwitz.setMaximumRelativeError(1.0e-10)
myAbdoRackwitz.setMaximumResidualError(1.0e-10)
myAbdoRackwitz.setMaximumConstraintError(1.0e-10)
myAbdoRackwitz.setStartingPoint(myDistribution.getMean())

myAlgoAR = ot.FORM(myAbdoRackwitz, myEvent)
myAlgoAR2 = ot.SORM(myAbdoRackwitz, myEvent)

myAlgoAR.run()
myAlgoAR2.run()

resultAR = myAlgoAR.getResult()
resultAR2 = myAlgoAR2.getResult()

#
# Monte Carlo
CoV_MC = 0.1
myMC = ot.ProbabilitySimulationAlgorithm(myEvent, ot.MonteCarloExperiment())
myMC.setMaximumOuterSampling(32000)
myMC.setBlockSize(1)
myMC.setMaximumCoefficientOfVariation(CoV_MC)
myMC.run()

#
# LHS
CoV_LHS = 0.1
myLHS = ot.ProbabilitySimulationAlgorithm(myEvent, ot.LHSExperiment())
myLHS.setMaximumOuterSampling(32000)
myLHS.setBlockSize(1)
myLHS.setMaximumCoefficientOfVariation(CoV_LHS)
myLHS.run()

#
#
# Results
#
#
# FORM/SORM Cobyla
PfC = resultC.getEventProbability()
Beta_generalizedC = resultC.getGeneralisedReliabilityIndex()
u_starC = resultC.getStandardSpaceDesignPoint()
x_starC = resultC.getPhysicalSpaceDesignPoint()
PtC = resultC.getIsStandardPointOriginInFailureSpace() and "true" or "false"
gammaC = resultC.getImportanceFactors()
beta_hasoferC = resultC.getHasoferReliabilityIndex()
SensitivityC = resultC.getEventProbabilitySensitivity()

PFBreitC2 = resultC2.getEventProbabilityBreitung()
BetaBreitC2 = resultC2.getGeneralisedReliabilityIndexBreitung()
PFHBC2 = resultC2.getEventProbabilityHohenbichler()
BetaHBC2 = resultC2.getGeneralisedReliabilityIndexHohenbichler()
PFTvedtC2 = resultC2.getEventProbabilityTvedt()
BetaTvedtC2 = resultC2.getGeneralisedReliabilityIndexTvedt()
CurvC2 = resultC2.getSortedCurvatures()
u_starC2 = resultC2.getStandardSpaceDesignPoint()
x_starC2 = resultC2.getPhysicalSpaceDesignPoint()
PtC2 = resultC2.getIsStandardPointOriginInFailureSpace() and "true" or "false"
gammaC2 = resultC2.getImportanceFactors()
beta_hasoferC2 = resultC2.getHasoferReliabilityIndex()

#
# FORM/SORM Abdo Rackwitz
PfAR = resultAR.getEventProbability()
Beta_generalizedAR = resultAR.getGeneralisedReliabilityIndex()
u_starAR = resultAR.getStandardSpaceDesignPoint()
x_starAR = resultAR.getPhysicalSpaceDesignPoint()
PtAR = resultAR.getIsStandardPointOriginInFailureSpace() and "true" or "false"
gammaAR = resultAR.getImportanceFactors()
beta_hasoferAR = resultAR.getHasoferReliabilityIndex()
SensitivityAR = resultAR.getEventProbabilitySensitivity()

PFBreitAR2 = resultAR2.getEventProbabilityBreitung()
BetaBreitAR2 = resultAR2.getGeneralisedReliabilityIndexBreitung()
PFHBAR2 = resultAR2.getEventProbabilityHohenbichler()
BetaHBAR2 = resultAR2.getGeneralisedReliabilityIndexHohenbichler()
PFTvedtAR2 = resultAR2.getEventProbabilityTvedt()
BetaTvedtAR2 = resultAR2.getGeneralisedReliabilityIndexTvedt()
CurvAR2 = resultAR2.getSortedCurvatures()
u_starAR2 = resultAR2.getStandardSpaceDesignPoint()
x_starAR2 = resultAR2.getPhysicalSpaceDesignPoint()
PtAR2 = resultAR2.getIsStandardPointOriginInFailureSpace() and "true" or "false"
gammaAR2 = resultAR2.getImportanceFactors()
beta_hasoferAR2 = resultAR2.getHasoferReliabilityIndex()

#
# Monte Carlo
ResultMC = myMC.getResult()
PFMC = ResultMC.getProbabilityEstimate()
CVMC = ResultMC.getCoefficientOfVariation()
Variance_PF_MC = ResultMC.getVarianceEstimate()
length90MC = ResultMC.getProbabilityDistribution().computeBilateralConfidenceInterval(
    0.9
)

#
# LHS
ResultLHS = myLHS.getResult()
PFLHS = ResultLHS.getProbabilityEstimate()
CVLHS = ResultLHS.getCoefficientOfVariation()
Variance_PF_LHS = ResultLHS.getVarianceEstimate()
length90LHS = ResultLHS.getProbabilityDistribution().computeBilateralConfidenceInterval(
    0.9
)

#

#
# Outputs
#
print("")
print("")
print(
    "************************************************************************************************"
)
print(
    "***************************************** FORM  COBYLA *****************************************"
)
print(
    "************************************************************************************************"
)
print("event probability = %.5e" % PfC)
print("generalized reliability index = %.5f" % Beta_generalizedC)
print(
    "************************************************************************************************"
)
for i in range(u_starC.getDimension()):
    print("standard space design point = %.4f" % u_starC[i])
print(
    "************************************************************************************************"
)
for i in range(x_starC.getDimension()):
    print("physical space design point = %.5f" % x_starC[i])
print(
    "************************************************************************************************"
)
print("is standard point origin in failure space? ", PtC)
print(
    "************************************************************************************************"
)
for i in range(gammaC.getDimension()):
    print("importance factors = %.5f" % gammaC[i])
print(
    "************************************************************************************************"
)
print("Hasofer reliability index = %.5f" % beta_hasoferC)
print(
    "************************************************************************************************"
)
for i in range(SensitivityC.getSize()):
    for j in range(SensitivityC[i].getDimension()):
        print("Pf sensitivity = %.5f" % SensitivityC[i][j])
print(
    "************************************************************************************************"
)
print("")
print(
    "************************************************************************************************"
)
print(
    "************************************** FORM ABDO RACKWITZ **************************************"
)
print(
    "************************************************************************************************"
)
print("event probability = %.5e" % PfAR)
print("generalized reliability index = %.5f" % Beta_generalizedAR)
print(
    "************************************************************************************************"
)
for i in range(u_starAR.getDimension()):
    print("standard space design point = %.4f" % u_starAR[i])
print(
    "************************************************************************************************"
)
for i in range(x_starAR.getDimension()):
    print("physical space design point = %.5f" % x_starAR[i])
print(
    "************************************************************************************************"
)
print("is standard point origin in failure space? ", PtAR)
print(
    "************************************************************************************************"
)
for i in range(gammaAR.getDimension()):
    print("importance factors = %.5f" % gammaAR[i])
print(
    "************************************************************************************************"
)
print("Hasofer reliability index = %.5f" % beta_hasoferAR)
print(
    "************************************************************************************************"
)
for i in range(SensitivityAR.getSize()):
    for j in range(SensitivityAR[i].getDimension()):
        print("Pf sensitivity = %.5f" % SensitivityAR[i][j])
print(
    "************************************************************************************************"
)
print("")
print(
    "************************************************************************************************"
)
print(
    "***************************************** SORM  COBYLA *****************************************"
)
print(
    "************************************************************************************************"
)
print("Breitung event probability = %.5e" % PFBreitC2)
print("Breitung generalized reliability index = %.5f" % BetaBreitC2)
print("Hohenbichler event probability = %.5e" % PFHBC2)
print("Hohenbichler generalized reliability index = %.5f" % BetaHBC2)
print("Tvedt event probability = %.5e" % PFTvedtC2)
print("Tvedt generalized reliability index = %.5f" % BetaTvedtC2)
print(
    "************************************************************************************************"
)
for i in range(CurvC2.getDimension()):
    print("sorted curvatures = %.5f" % cleanScalar(CurvC2[i]))
print(
    "************************************************************************************************"
)
for i in range(u_starC2.getDimension()):
    print("standard space design point = %.4f" % u_starC2[i])
print(
    "************************************************************************************************"
)
for i in range(x_starC2.getDimension()):
    print("physical space design point = %.5f" % x_starC2[i])
print(
    "************************************************************************************************"
)
print(
    "************************************************************************************************"
)
print("is standard point origin in failure space? ", PtC2)
print(
    "************************************************************************************************"
)
for i in range(gammaC2.getDimension()):
    print("importance factors = %.5f" % gammaC2[i])
print(
    "************************************************************************************************"
)
print("Hasofer reliability index = %.5f" % beta_hasoferC2)
print(
    "************************************************************************************************"
)
print("")
print(
    "************************************************************************************************"
)
print(
    "************************************** SORM ABDO RACKWITZ **************************************"
)
print(
    "************************************************************************************************"
)
print("Breitung event probability = %.5e" % PFBreitAR2)
print("Breitung generalized reliability index = %.5f" % BetaBreitAR2)
print("Hohenbichler event probability = %.5e" % PFHBAR2)
print("Hohenbichler generalized reliability index = %.5f" % BetaHBAR2)
print("Tvedt event probability = %.5e" % PFTvedtAR2)
print("Tvedt generalized reliability index = %.5f" % BetaTvedtAR2)
print(
    "************************************************************************************************"
)
for i in range(CurvAR2.getDimension()):
    print("sorted curvatures = %.5f" % cleanScalar(CurvAR2[i]))
print(
    "************************************************************************************************"
)
for i in range(u_starAR2.getDimension()):
    print("standard space design point = %.4f" % u_starAR2[i])
print(
    "************************************************************************************************"
)
for i in range(x_starAR2.getDimension()):
    print("physical space design point = %.5f" % x_starAR2[i])
print(
    "************************************************************************************************"
)
print(
    "************************************************************************************************"
)
print("is standard point origin in failure space? ", PtAR2)
print(
    "************************************************************************************************"
)
for i in range(gammaAR2.getDimension()):
    print("importance factors = %.5f" % gammaAR2[i])
print(
    "************************************************************************************************"
)
print("Hasofer reliability index = %.5f" % beta_hasoferAR2)
print(
    "************************************************************************************************"
)
print("")
print(
    "************************************************************************************************"
)
print(
    "**************************************** MONTE CARLO *******************************************"
)
print(
    "************************************************************************************************"
)
print("Pf estimation = %.5e" % PFMC)
print("Pf Variance estimation = %.5e" % Variance_PF_MC)
print("CoV = %.5f" % CVMC)
print("90% Confidence Interval =", length90MC)
print(
    "************************************************************************************************"
)
print("")
print(
    "************************************************************************************************"
)
print(
    "******************************************* L H S **********************************************"
)
print(
    "************************************************************************************************"
)
print("Pf estimation = %.5e" % PFLHS)
print("Pf Variance estimation = %.5e" % Variance_PF_LHS)
print("CoV = %.5f" % CVLHS)
print("90% Confidence Interval =", length90LHS)
print(
    "************************************************************************************************"
)
print("")



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# linear
levelFunction = ot.SymbolicFunction(["x1", "x2", "x3", "x4"], ["x1+2*x2-3*x3+4*x4"])
bounds = ot.Interval([-3.0] * 4, [5.0] * 4)
algo = ot.TNC()
algo.setStartingPoint([0.0] * 4)

problem = ot.OptimizationProblem(levelFunction)
problem.setBounds(bounds)
problem.setMinimization(True)

algo.setProblem(problem)
print("algo=", algo)
algo.run()
result = algo.getResult()
print("result=", result.getOptimalPoint())
print("multipliers=", result.computeLagrangeMultipliers())

problem.setMinimization(False)
algo.setProblem(problem)
print("algo=", algo)
algo.run()
result = algo.getResult()
print("result=", result.getOptimalPoint())
print("multipliers=", result.computeLagrangeMultipliers())

# non-linear
levelFunction = ot.SymbolicFunction(
    ["x1", "x2", "x3", "x4"], ["x1*cos(x1)+2*x2*x3-3*x3+4*x3*x4"]
)

bounds = ot.Interval(ot.Point(4, -3.0), ot.Point(4, 5.0))

algo = ot.TNC()
problem = ot.OptimizationProblem(levelFunction)
problem.setBounds(bounds)

problem.setMinimization(True)
algo.setProblem(problem)
algo.setStartingPoint([3.0, -2.5, 4.5, -2.5])
print("algo=", algo)
algo.run()
result = algo.getResult()
print("result=", result.getOptimalPoint())
print("multipliers=", result.computeLagrangeMultipliers())

problem.setMinimization(False)
algo.setProblem(problem)
print("algo=", algo)
algo.setStartingPoint([-2.5, 4.5, 4.5, 4.5])
algo.run()
result = algo.getResult()
print("result=", result.getOptimalPoint())
print("multipliers=", result.computeLagrangeMultipliers())



#! /usr/bin/env python

import openturns as ot
import openturns.experimental as otexp
import openturns.testing as ott

ot.TESTPREAMBLE()

# inspired from highs example: https://github.com/ERGO-Code/HiGHS/blob/master/examples/call_highs_from_cpp.cpp
# min f = 1.1*x_0 + x_1
# ineq:
#        x_1 <= 7
#   5 <= x_0 + 2x_1 <= 15
#   6 <= 3x_0 + 2x_1
# bounds:
#   0 <= x_0 <= 4; 1 <= x_1
ot.ResourceMap.AddAsBool("HiGHS-output_flag", True)
bounds = ot.Interval([0.0, 1.0], [4.0, 1e30])
cost = [1.1, 1.0]
A = ot.Matrix([[0.0, 1.0], [1.0, 2.0], [3.0, 2.0]])
cb = ot.Interval([-1e9, 5.0, 6.0], [7.0, 15.0, 1e9])
problem = otexp.LinearProblem(cost, bounds, A, cb)
print(problem)

sol = {
    ot.OptimizationProblemImplementation.CONTINUOUS: [0.5, 2.25],
    ot.OptimizationProblemImplementation.INTEGER: [0.0, 3.0],
}
for vtype in [
    ot.OptimizationProblemImplementation.CONTINUOUS,
    ot.OptimizationProblemImplementation.INTEGER,
]:
    problem.setVariablesType([vtype] * 2)
    algo = otexp.HiGHS(problem)
    print(algo)
    algo.run()
    result = algo.getResult()
    print(result)
    x = result.getOptimalPoint()
    ott.assert_almost_equal(x, sol[vtype])



"""
Multi-objective optimization using Pagmo
========================================
"""

# %%
# In this example we are going to explore optimization using the :class:`~openturns.Pagmo` solver.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# List available algorithms
for algo in ot.Pagmo.GetAlgorithmNames():
    print(algo)

# %%
# Create the problem, from ZDT test suite
f = ot.SymbolicFunction(
    ["x1", "x2"], ["x1", "var g := 1.0 + 9.0 * (x1 + x2); g * (1.0 - sqrt(x1 / g))"]
)
zdt1 = ot.OptimizationProblem(f)
zdt1.setBounds(ot.Interval([0.0] * 2, [1.0] * 2))

# %%
# We create the first generation of points by sampling into the bounding box
pop0 = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 2).getSample(100)

# %%
# We create the algorithm that should evolve over 10 generations
algo = ot.Pagmo(zdt1, "nsga2", pop0)
algo.setMaximumIterationNumber(10)

# %%
# Benefit from parallel evaluations if the function allows it
algo.setBlockSize(8)

# %%
# We run the algo
algo.run()
pop1 = algo.getResult().getFinalPoints()

# %%
# We list the available Pareto fronts
fronts = algo.getResult().getParetoFrontsIndices()
len(fronts)

# %%
# We show the Pareto front
graph = ot.Graph("Pareto front", "y1", "y2")
graph.setLegendPosition("upper right")
front = algo.getResult().getFinalValues().select(fronts[0]).sortAccordingToAComponent(0)
data = ot.Sample(2 * front.getSize() - 1, 2)
for i in range(front.getSize()):
    data[2 * i] = front[i]
    if i != front.getSize() - 1:
        data[2 * i + 1, 0] = front[i + 1, 0]
        data[2 * i + 1, 1] = front[i, 1]
curve = ot.Curve(data)
curve.setLegend(f"front {0}")
graph.add(curve)
graph.setGrid(True)
_ = otv.View(graph)

# %%
# We show the Pareto front from successive generations
fronts = []
for gen in range(5):
    algo = ot.Pagmo(zdt1, "nsga2", pop0)
    algo.setMaximumIterationNumber(gen)
    algo.run()
    front0 = algo.getResult().getParetoFrontsIndices()[0]
    fronts.append(algo.getResult().getFinalValues().select(front0))
graph = ot.Graph("Successive fronts", "y1", "y2")
graph.setLegendPosition("upper right")
for k in range(len(fronts)):
    front = fronts[k].sortAccordingToAComponent(0)
    print(front)
    data = ot.Sample(2 * front.getSize() - 1, 2)
    for i in range(front.getSize()):
        data[2 * i] = front[i]
        if i != front.getSize() - 1:
            data[2 * i + 1, 0] = front[i + 1, 0]
            data[2 * i + 1, 1] = front[i, 1]
    curve = ot.Curve(data)
    curve.setLegend(f"generation {k}")
    graph.add(curve)
graph.setGrid(True)
_ = otv.View(graph)


otv.View.ShowAll()

# %%



#! /usr/bin/env python

import openturns as ot
from openturns.testing import assert_almost_equal
import openturns.experimental as otexp

ot.TESTPREAMBLE()

print("Fit y ~ 3 - 2 x1 + x2 + epsilon")
kFoldParameter = 4
foldRootSize = 3
# Makes so that k does not divide the sample size.
# In this case, we must take the different weigths
# of each fold into account.
sampleSize = foldRootSize * kFoldParameter + 1
print("sampleSize = ", sampleSize)
aCollection = []
marginal1 = ot.Uniform(-1.0, 1.0)
aCollection.append(marginal1)
aCollection.append(marginal1)
distribution = ot.JointDistribution(aCollection)
inputSample = distribution.getSample(sampleSize)
print("inputSample=", inputSample)
g = ot.SymbolicFunction(["x1", "x2"], ["3 - 2 * x1 + x2"])
noise = ot.Normal(0.0, 0.5)
outputSample = g(inputSample) + noise.getSample(sampleSize)
print("outputSample=", outputSample)
lmAlgo = ot.LinearModelAlgorithm(inputSample, outputSample)
result = lmAlgo.getResult()

# Create LOO validation
splitterLOO = ot.LeaveOneOutSplitter(sampleSize)
validationLOO = otexp.LinearModelValidation(result, splitterLOO)
print(validationLOO)

# Compute analytical LOO MSE
print("Compute Analytical LOO MSE")
mseLOOAnalytical = validationLOO.computeMeanSquaredError()
print("Analytical LOO MSE =", mseLOOAnalytical[0])

# Compute naive leave-one-out
residualsLOO = ot.Point(sampleSize)
j = 0
for indicesTrain, indicesTest in splitterLOO:
    inputSampleTrainLOO = inputSample[indicesTrain]
    inputSampleLOOTest = inputSample[indicesTest]
    outputSampleTrainLOO = outputSample[indicesTrain]
    outputSampleLOOTest = outputSample[indicesTest]
    lmAlgoLOO = ot.LinearModelAlgorithm(inputSampleTrainLOO, outputSampleTrainLOO)
    resultLOO = lmAlgoLOO.getResult()
    metamodelLOO = resultLOO.getMetaModel()
    predictionLOOTest = metamodelLOO(inputSampleLOOTest)
    residualsLOOTest = predictionLOOTest.asPoint() - outputSampleLOOTest.asPoint()
    residualsLOO[j] = residualsLOOTest[0]
    j += 1

mseLOOnaive = residualsLOO.normSquare() / sampleSize
print("Naive LOO MSE      =", mseLOOnaive)

# Test
rtolLOO = 1.0e-12
atolLOO = 0.0
assert_almost_equal(mseLOOAnalytical[0], mseLOOnaive, rtolLOO, atolLOO)

# Check LOO R2
r2ScoreLOO = validationLOO.computeR2Score()
print("Analytical LOO R2 score = ", r2ScoreLOO[0])
sampleVariance = outputSample.computeCentralMoment(2)
r2ScoreReference = 1.0 - mseLOOAnalytical[0] / sampleVariance[0]
print("Computed R2 score       = ", r2ScoreReference)
rtolLOO = 1.0e-12
atolLOO = 0.0
assert_almost_equal(r2ScoreReference, r2ScoreLOO[0], rtolLOO, atolLOO)

# Create KFold validation
splitterKFold = ot.KFoldSplitter(sampleSize, kFoldParameter)
validationKFold = otexp.LinearModelValidation(result, splitterKFold)
print(validationKFold)

# Compute analytical KFold MSE
mseKFoldAnalytical = validationKFold.computeMeanSquaredError()
print("Analytical KFold MSE =", mseKFoldAnalytical[0])

# Naive KFold
residualsKFold = ot.Point(sampleSize)
foldIndex = 0
for indicesTrain, indicesTest in splitterKFold:
    inputSampleKFoldTrain = inputSample[indicesTrain]
    outputSampleKFoldTrain = outputSample[indicesTrain]
    inputSampleKFoldTest = inputSample[indicesTest]
    outputSampleKFoldTest = outputSample[indicesTest]
    lmAlgoKFold = ot.LinearModelAlgorithm(inputSampleKFoldTrain, outputSampleKFoldTrain)
    resultKFold = lmAlgoKFold.getResult()
    metamodelKFold = resultKFold.getMetaModel()
    predictionKFoldTest = metamodelKFold(inputSampleKFoldTest)
    residualsKFoldTest = predictionKFoldTest.asPoint() - outputSampleKFoldTest.asPoint()
    foldSize = indicesTest.getSize()
    for k in range(foldSize):
        residualsKFold[indicesTest[k]] = residualsKFoldTest[k]
    foldIndex += 1

mseKFoldnaive = residualsKFold.normSquare() / sampleSize
print("Naive KFold MSE      =", mseKFoldnaive)

# Test
rtolKFold = 1.0e-7
atolKFold = 0.0
assert_almost_equal(mseKFoldAnalytical[0], mseKFoldnaive, rtolKFold, atolKFold)

# Check K-Fold R2
r2ScoreKFold = validationKFold.computeR2Score()
print("Analytical K-Fold R2 score =", r2ScoreKFold[0])
r2ScoreReference = 1.0 - mseKFoldAnalytical[0] / sampleVariance[0]
print("Computed R2 score          =", r2ScoreReference)
rtolKFold = 1.0e-12
atolKFold = 0.0
assert_almost_equal(r2ScoreReference, r2ScoreKFold[0], rtolLOO, atolLOO)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Create a KarhunenLoeveResult
mesh = ot.IntervalMesher([9]).build(ot.Interval(-1.0, 1.0))
cov1D = ot.AbsoluteExponential([1.0])
algo = ot.KarhunenLoeveP1Algorithm(mesh, cov1D, 0.0)
algo.run()
result = algo.getResult()
lifting = ot.KarhunenLoeveLifting(result)
# Construction based on a Function followed by a PointToFieldFunction
dimension = lifting.getInputDimension()
function = ot.SymbolicFunction(
    ot.Description.BuildDefault(dimension, "x"),
    ot.Description.BuildDefault(dimension, "x"),
)
# Create an instance
myFunc = ot.PointToFieldConnection(lifting, function)

print("myFunc=", myFunc)
# Get the input and output description
print("myFunc input description=", myFunc.getInputDescription())
print("myFunc output description=", myFunc.getOutputDescription())
# Get the input and output dimension
print("myFunc input dimension=", myFunc.getInputDimension())
print("myFunc output dimension=", myFunc.getOutputDimension())
# Connection on a point
point = [1.0] * myFunc.getInputDimension()
print("point=", point)
print("myFunc(point)=", myFunc(point))
# Get the number of calls
print("called ", myFunc.getCallsNumber(), " times")

# Construction based on a PointToFieldFunction followed by a
# FieldToPointFunction
fieldFunction = ot.ValueFunction(ot.SymbolicFunction("x", "x"), mesh)
# Create an instance
myFunc = ot.PointToFieldConnection(fieldFunction, lifting)

print("myFunc=", myFunc)
# Get the input and output description
print("myFunc input description=", myFunc.getInputDescription())
print("myFunc output description=", myFunc.getOutputDescription())
# Get the input and output dimension
print("myFunc input dimension=", myFunc.getInputDimension())
print("myFunc output dimension=", myFunc.getOutputDimension())
# Connection on a point
point = [1.0] * myFunc.getInputDimension()
print("point=", point)
print("myFunc(point)=", myFunc(point))
# Get the number of calls
print("called ", myFunc.getCallsNumber(), " times")



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# start and step
start = -2.0
step = 0.1
# Number of steps for the RegularGrid
n = 10
regularGrid = ot.RegularGrid(start, step, n)

print("regularGrid =", regularGrid)
print("Start time =", regularGrid.getStart())
print("End time =", regularGrid.getEnd())
print("Time step =", regularGrid.getStep())
print("Steps =", regularGrid.getN())



#! /usr/bin/env python

import openturns as ot
from openturns.usecases import flood_model

ot.TESTPREAMBLE()

sampleSize = 500
totalDegree = 7

# Test pretty-print with small dimension output
fm = flood_model.FloodModel()
inputDescription = fm.model.getInputDescription()
marginals = [fm.distribution.getMarginal(i) for i in range(fm.dim)]
basis = ot.OrthogonalProductPolynomialFactory(marginals)
inputSample = fm.distribution.getSample(sampleSize)
outputSample = fm.model(inputSample)
selectionAlgorithm = ot.LeastSquaresMetaModelSelectionFactory()
projectionStrategy = ot.LeastSquaresStrategy(selectionAlgorithm)
enumerateFunction = basis.getEnumerateFunction()
basisSize = enumerateFunction.getBasisSizeFromTotalDegree(totalDegree)
adaptiveStrategy = ot.FixedStrategy(basis, basisSize)
algo = ot.FunctionalChaosAlgorithm(
    inputSample, outputSample, fm.distribution, adaptiveStrategy, projectionStrategy
)
algo.run()
chaosResult = algo.getResult()
chaosSobol = ot.FunctionalChaosSobolIndices(chaosResult)
print(chaosSobol)
print(chaosSobol._repr_html_())
print(chaosSobol.__repr_markdown__())


class RepeatedFloodOutputDimensionFunction(ot.OpenTURNSPythonFunction):
    def __init__(self, outputDimension):
        super().__init__(4, outputDimension)
        self.fm = flood_model.FloodModel()
        self.outputDimension = outputDimension

    def _exec(self, x):
        y = ot.Point(self.outputDimension)
        y_base = fm.model(x)
        for i in range(self.outputDimension):
            y[i] = i + y_base[0]
        return y


# Test pretty-print with large dimension output
outputDimension = 5
model = ot.Function(RepeatedFloodOutputDimensionFunction(outputDimension))
sampleSize = 500
totalDegree = 7
fm = flood_model.FloodModel()
inputDescription = fm.model.getInputDescription()
marginals = [fm.distribution.getMarginal(i) for i in range(fm.dim)]
basis = ot.OrthogonalProductPolynomialFactory(marginals)
inputSample = fm.distribution.getSample(sampleSize)
outputSample = model(inputSample)
selectionAlgorithm = ot.LeastSquaresMetaModelSelectionFactory()
projectionStrategy = ot.LeastSquaresStrategy(selectionAlgorithm)
enumerateFunction = basis.getEnumerateFunction()
basisSize = enumerateFunction.getBasisSizeFromTotalDegree(totalDegree)
adaptiveStrategy = ot.FixedStrategy(basis, basisSize)
chaosAlgo = ot.FunctionalChaosAlgorithm(
    inputSample, outputSample, fm.distribution, adaptiveStrategy, projectionStrategy
)
chaosAlgo.run()
chaosResult = chaosAlgo.getResult()
chaosSobol = ot.FunctionalChaosSobolIndices(chaosResult)
print(chaosSobol)
print(chaosSobol._repr_html_())
print(chaosSobol.__repr_markdown__())



"""
Estimate extrema iteratively
============================
"""

# %%
# In this example, we compute extrema iteratively.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# We first create a one-dimensional Uniform random variable to generate data.
dim = 1
distNormal = ot.Uniform()

# %%
# The :class:`~openturns.IterativeExtrema` class needs the dimension of the sample
# (here 1):

# %%
iterExtrema = ot.IterativeExtrema(dim)

# %%
# We can now perform the simulations.
# In our case most of the data should be in the [-3,3] interval.
# Consequently with few samples the expected minimum should be around -3
# and the expected maximum should be around 3.
# We first increment the object with one :class:`~openturns.Point` at a time.
# At any given step the current minimum is obtained thanks to
# the :meth:`~openturns.IterativeExtrema.getMin` method, the current maximum
# with the :meth:`~openturns.IterativeExtrema.getMax` method and the
# current number of iterations is given by the
# :meth:`~openturns.IterativeMoments.getIterationNumber` method.

# %%
size = 2000
minEvolution = ot.Sample()
maxEvolution = ot.Sample()
for i in range(size):
    point = distNormal.getRealization()
    iterExtrema.increment(point)
    minEvolution.add(iterExtrema.getMin())
    maxEvolution.add(iterExtrema.getMax())

# %%
# We display the evolution of the minimum (in blue) and the maximum (orange).
iterationSample = ot.Sample.BuildFromPoint(range(1, size + 1))
#
curveMin = ot.Curve(iterationSample, minEvolution)
curveMin.setLegend("min.")
#
curveMax = ot.Curve(iterationSample, maxEvolution)
curveMax.setLegend("max.")
#
graph = ot.Graph("Evolution of the min/max", "iteration nb", "min/max")
graph.add(curveMin)
graph.add(curveMax)
graph.setLegendPosition("upper left")
graph.setLogScale(ot.GraphImplementation.LOGX)
view = otv.View(graph)

# %%
# We can also increment with a :class:`~openturns.Sample`.

# %%
sample = distNormal.getSample(size)
iterExtrema.increment(sample)

# %%
# We print the total number of iterations and the extrema.
print("Total number of iterations: " + str(iterExtrema.getIterationNumber()))
print("Minimum: ", iterExtrema.getMin())
print("Maximum: ", iterExtrema.getMax())

otv.View.ShowAll()



"""
Distribution manipulation
=========================
"""

# %%
# In this example we are going to exhibit some of the services exposed by the distribution objects:
#
# - ask for the dimension, with the method `getDimension` ;
# - extract the marginal distributions, with the method `getMarginal` ;
# - to ask for some properties, with `isContinuous`, `isDiscrete`, `isElliptical` ;
# - to get the copula, with the method `getCopula` ;
# - to ask for some properties on the copula, with the methods `hasIndependentCopula`, `hasEllipticalCopula` ;
# - to evaluate some moments, with `getMean`, `getStandardDeviation`, `getCovariance`, `getSkewness`, `getKurtosis` ;
# - to evaluate the roughness, with the method `getRoughness` ;
# - to get one realization or simultaneously :math:`n` realizations, with the method `getRealization`, `getSample` ;
# - to evaluate the probability content of a given interval, with the method `computeProbability` ;
# - to evaluate a quantile or a complementary quantile, with the method `computeQuantile` ;
# - to evaluate the characteristic function of the distribution ;
# - to evaluate the derivative of the CDF or PDF ;
# - to draw some curves.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Create an 1-d distribution
dist_1 = ot.Normal()

# Create a 2-d distribution
dist_2 = ot.JointDistribution(
    [ot.Normal(), ot.Triangular(0.0, 2.0, 3.0)], ot.ClaytonCopula(2.3)
)

# Create a 3-d distribution
copula_dim3 = ot.Student(5.0, 3).getCopula()
dist_3 = ot.JointDistribution(
    [ot.Normal(), ot.Triangular(0.0, 2.0, 3.0), ot.Exponential(0.2)], copula_dim3
)

# %%
# Get the dimension fo the distribution
dist_2.getDimension()

# %%
# Get the 2nd marginal
dist_2.getMarginal(1)

# %%
# Get a 2-d marginal
dist_3.getMarginal([0, 1]).getDimension()

# %%
# Ask some properties of the distribution
dist_1.isContinuous(), dist_1.isDiscrete(), dist_1.isElliptical()

# %%
# Get the copula
copula = dist_2.getCopula()

# %%
# Ask some properties on the copula
dist_2.hasIndependentCopula(), dist_2.hasEllipticalCopula()

# %%
# Get the mean vector of the distribution
dist_2.getMean()

# %%
# Get the standard deviation vector of the distribution
dist_2.getStandardDeviation()

# %%
# Get the covariance matrix of the distribution
dist_2.getCovariance()

# %%
# Get the skewness vector of the distribution
dist_2.getSkewness()

# %%
# Get the kurtosis vector of the distribution
dist_2.getKurtosis()

# %%
# Get the roughness of the distribution
dist_1.getRoughness()

# %%
# Get one realization
dist_2.getRealization()

# %%
# Get several realizations
dist_2.getSample(5)

# %%
# Evaluate the PDF at the mean point
dist_2.computePDF(dist_2.getMean())

# %%
# Evaluate the CDF at the mean point
dist_2.computeCDF(dist_2.getMean())

# %%
# Evaluate the complementary CDF
dist_2.computeComplementaryCDF(dist_2.getMean())

# %%
# Evaluate the survival function at the mean point
dist_2.computeSurvivalFunction(dist_2.getMean())

# %%
# Evaluate the PDF on a sample
dist_2.computePDF(dist_2.getSample(5))

# %%
# Evaluate the CDF on a sample
dist_2.computeCDF(dist_2.getSample(5))

# %%
# Evaluate the probability content of an 1-d interval
interval = ot.Interval(-2.0, 3.0)
dist_1.computeProbability(interval)

# %%
# Evaluate the probability content of a 2-d interval
interval = ot.Interval([0.4, -1], [3.4, 2])
dist_2.computeProbability(interval)

# %%
# Evaluate the quantile of order `p=90%`
dist_2.computeQuantile(0.90)

# %%
# and the quantile of order 1-p
dist_2.computeQuantile(0.90, True)

# %%
# Evaluate the quantiles of order p et q
# For example, the quantile `90%` and `95%`
dist_1.computeQuantile([0.90, 0.95])

# %%
# and the quantile of order 1-p and 1-q
dist_1.computeQuantile([0.90, 0.95], True)

# %%
# Evaluate the characteristic function of the distribution (only 1-d)
dist_1.computeCharacteristicFunction(dist_1.getMean()[0])

# %%
# Evaluate the derivatives of the PDF with respect to the parameters at mean
dist_2.computePDFGradient(dist_2.getMean())

# %%
# Evaluate the derivatives of the CDF with respect to the parameters at mean
dist_2.computeCDFGradient(dist_2.getMean())

# %%
# Draw PDF
graph = dist_1.drawPDF()
view = otv.View(graph)

# %%
# Draw CDF
graph = dist_1.drawCDF()
view = otv.View(graph)

# %%
# Draw an 1-d quantile curve

# Define the range and the number of points
qMin = 0.2
qMax = 0.6
nbrPoints = 101
quantileGraph = dist_1.drawQuantile(qMin, qMax, nbrPoints)
view = otv.View(quantileGraph)

# %%
# Draw a 2-d quantile curve

# Define the range and the number of points
qMin = 0.3
qMax = 0.9
nbrPoints = 101
quantileGraph = dist_2.drawQuantile(qMin, qMax, nbrPoints)
view = otv.View(quantileGraph)

# %%
# Display all figures
otv.View.ShowAll()



"""
Estimate Sobol indices on a field to point function
===================================================
"""

# %%
# In this example, we are going to perform sensitivity analysis of an application that takes
# fields as input and vectors as output from a sample of data:
#
# .. math::
#     h: \left|
#       \begin{array}{ccl}
#          \cM_N \times (\Rset^d)^N & \rightarrow & \Rset^p \\
#          \mat{X} & \mapsto & \vect{Y}
#       \end{array}
#     \right.
#
# This involves these steps:
#
# - Generate some input/output data matching the application :math:`h`
# - Run the :class:`~openturns.FieldToPointFunctionalChaosAlgorithm` class
# - Validate the Karhunen-Loeve decompositions of the inputs
# - Validate the chaos metamodel between the KL coefficients and the outputs
# - Retrieve the Sobol' indices from :class:`openturns.FieldFunctionalChaosSobolIndices`
#

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# First build a process to generate the input data.
# We assemble a 4-d process from functional and Gaussian processes.
T = 3.0
NT = 32
tg = ot.RegularGrid(0.0, T / NT, NT)
f1 = ot.SymbolicFunction(["t"], ["sin(t)"])
f2 = ot.SymbolicFunction(["t"], ["cos(t)^2"])
coeff1_dist = ot.Normal([1.0] * 2, [0.6] * 2, ot.CorrelationMatrix(2))
p1 = ot.FunctionalBasisProcess(coeff1_dist, ot.Basis([f1, f2]), tg)
p2 = ot.GaussianProcess(ot.SquaredExponential([1.0], [T / 4.0]), tg)
coeff3_dist = ot.JointDistribution([ot.Uniform(), ot.Normal()])
f1 = ot.SymbolicFunction(["t"], ["1", "0"])
f2 = ot.SymbolicFunction(["t"], ["0", "1"])
p3 = ot.FunctionalBasisProcess(coeff3_dist, ot.Basis([f1, f2]))
X = ot.AggregatedProcess([p1, p2, p3])
X.setMesh(tg)

# %%
# Draw some input trajectories from our process
x = X.getSample(10)
graph = x.drawMarginal(0)
graph.setTitle(f"{x.getSize()} input trajectories")
_ = otv.View(graph)

# %%
# Generate input realizations and the corresponding output from a Field->Point function


class pyf2p(ot.OpenTURNSPythonFieldToPointFunction):
    def __init__(self, mesh):
        super(pyf2p, self).__init__(mesh, 4, 1)
        self.setInputDescription(["x1", "x2", "x3", "x4"])
        self.setOutputDescription(["y"])

    def _exec(self, X):
        Xs = ot.Sample(X)
        x1, x2, x3, x4 = Xs.computeMean()
        y = x1 + x2 + x3 - x4 + x1 * x2 - x3 * x4 - 0.1 * x1 * x2 * x3
        return [y]


f = ot.FieldToPointFunction(pyf2p(tg))
N = 1000
x = X.getSample(N)
y = f(x)

# %%
# Run the field-vector algorithm that performs KL-decomposition of the inputs
# and chaos learning between the KL coefficients and the output vectors
algo = ot.FieldToPointFunctionalChaosAlgorithm(x, y)
# 1. KL parameters
algo.setCenteredSample(False)  # our input sample is not centered (default)
algo.setThreshold(4e-2)  # we expect to explain 96% of variance
algo.setRecompress(
    False
)  # whether to re-truncate modes according to a global eigen value threshold across inputs (default)
algo.setNbModes(10)  # max KL modes (default=unlimited)
# 2. chaos parameters:
bs = ot.ResourceMap.GetAsUnsignedInteger("FunctionalChaosAlgorithm-BasisSize")
ot.ResourceMap.SetAsUnsignedInteger(
    "FunctionalChaosAlgorithm-BasisSize", N
)  # chaos basis size
ot.ResourceMap.SetAsBool("FunctionalChaosAlgorithm-Sparse", True)
algo.run()
ot.ResourceMap.SetAsUnsignedInteger("FunctionalChaosAlgorithm-BasisSize", bs)
result = algo.getResult()

# %%
# Retrieve the eigen values of each KL decomposition:
# we observe that each input process is represented by a different number of modes.
kl_results = result.getInputKLResultCollection()
n_modes = [len(res.getEigenvalues()) for res in kl_results]
print(f"n_modes={n_modes}")

# %%
# Retrieve the ratios of selected variance over cumulated variance:
# we see that all 3 inputs are perfectly represented, and the 2nd input almost perfectly.
ratios = [res.getSelectionRatio() for res in kl_results]
print(f"ratios={ratios}")

# %%
# Graphically validate the KL decompositions:
# we also see that the 2nd input appear to be less well represented than the others.
# Note however that the 0.98 selected/cumulated variance ratio actually means it is very good.
graphs = []
for i in range(x.getDimension()):
    validation = ot.KarhunenLoeveValidation(x.getMarginal(i), kl_results[i])
    graph = validation.drawValidation().getGraph(0, 0)
    graph.setTitle(f"KL validation - marginal #{i} ratio={100.0 * ratios[i]:.2f} %")
    otv.View(graph)
    graphs.append(graph)

# %%
# On the 2nd marginal we can filter out the points inside the 99% level-set
# to see that actually only a few points out of N are actually outliers.
graph = graphs[1]
data = graph.getDrawable(1).getData()
normal = ot.NormalFactory().build(data)
log_pdf = normal.computeLogPDF(data).asPoint()
l_pair = [(log_pdf[i], data[i]) for i in range(len(data))]
l_pair.sort(key=lambda t: t[0])
index_bad = int(0.01 * len(data))  # here 0.01 = (100-99)%
beta = l_pair[index_bad][0]
gnorm = normal.drawLogPDF(data.getMin(), data.getMax())
bad = [l_pair[i][1] for i in range(index_bad + 1)]
c = ot.Cloud(bad)
c.setPointStyle("bullet")
graph.setDrawable(1, c)
dr = gnorm.getDrawable(0)
dr.setLevels([beta])
dr.setLegend("99% level-set")
graph.add(dr)
_ = otv.View(graph)

# %%
# Graphically validate the chaos result:
# we can see the points are very close to the diagonal; this means
# approximated points are very close to the learning points.
modes = result.getModesSample()
metamodel = result.getFCEResult().getMetaModel()
output = result.getOutputSample()
metamodelPredictions = metamodel(modes)
validation = ot.MetaModelValidation(output, metamodelPredictions)
r2Score = validation.computeR2Score()
print(f"R2={r2Score}")
graph = validation.drawValidation()
graph.setTitle(f"Chaos validation - R2={r2Score}")
_ = otv.View(graph)

# %%
# Perform an evaluation on a new realization and ensure the output
# is close to the evaluation with the reference function
metamodel = result.getFieldToPointMetaModel()
x0 = X.getRealization()
y0 = f(x0)
y0hat = metamodel(x0)
print(f"y0={y0} y0^={y0hat}")

# %%
# Retrieve the first order Sobol' indices
# The preponderant variables are x2, x4 whereas x1, x3 have a low influence on the output
sensitivity = ot.FieldFunctionalChaosSobolIndices(result)
sobol_0 = sensitivity.getFirstOrderIndices()
print(f"first order={sobol_0}")

# %%
# Retrieve the total Sorder obol' indices
# The x3,x4 variables have total order indices significantly different than
# their first order indices counterpart meaning they interact with other variables
sobol_0t = sensitivity.getTotalOrderIndices()
print(f"total order={sobol_0t}")

# %%
# Draw the Sobol' indices
graph = sensitivity.draw()
view = otv.View(graph)

# %%
otv.View.ShowAll()



"""
Use case : flood model
======================
"""

import openturns as ot


class FloodModel:
    r"""
    Data class for the flood model.

    Parameters
    ----------

    L : float, optional
        Length of the river. The default is 5000.0.

    B : float, optional
        Width of the river. The default is 300.0.

    trueKs : float, optional
        The true value of the Ks parameter. The default is 30.0.

    trueZv : float, optional
        The true value of the Zv parameter. The default is 50.0.

    trueZm : float, optional
        The true value of the Zm parameter. The default is 55.0.

    distributionHdLow : bool, optional
        If True, then the distribution of Hd is uniform in [2, 4] i.e
        the dyke is relatively low.
        Otherwise, the distribution of Hd is uniform in [7, 9] i.e
        the dyke is relatively high.
        The default is True.

    Attributes
    ----------

    dim : int
        The dimension of the problem: dim=4

    Q : :class:`~openturns.TruncatedDistribution`
        `ot.TruncatedDistribution(ot.Gumbel(558.0, 1013.0), 0.0, ot.TruncatedDistribution.LOWER)`

    Ks : :class:`~openturns.TruncatedDistribution`
        `ot.TruncatedDistribution(ot.Normal(30.0, 7.5), 0.0, ot.TruncatedDistribution.LOWER)`

    Zv : :class:`~openturns.Uniform`
        `ot.Uniform(49.0, 51.0)`

    Zm : :class:`~openturns.Uniform`
        `ot.Uniform(54.0, 56.0)`

    B : :class:`~openturns.Uniform`
        `Triangular(295.0, 300.0, 305.0)`

    L : :class:`~openturns.Uniform`
        `ot.Triangular(4990.0, 5000.0, 5010.0)`

    Hd : :class:`~openturns.Uniform`
        `ot.Uniform(54.0, 56.0)`

    Zb : :class:`~openturns.Uniform`
        The distribution depends on `distributionHdLow`.

    model : :class:`~openturns.ParametricFunction`
        The flood model.
        The function has input dimension 8 and output dimension 3.
        More precisely, we have  :math:`\vect{X} = (Q, K_s, Z_v, Z_m, B, L, Z_b, H_d)` and
        :math:`\vect{Y} = (H, S, C)`.

    distribution : :class:`~openturns.JointDistribution`
        The joint distribution of the input parameters.

    data : :class:`~openturns.Sample`
        of size 10 and dimension 2
        A data set which contains noisy observations of the flow rate (column 0)
        and the height (column 1).

    Examples
    --------
    >>> from openturns.usecases import flood_model
    >>> # Load the flood model
    >>> fm = flood_model.FloodModel()
    >>> print(fm.data[:5])
        [ Q ($m^3/s$) H (m)       ]
    0 : [  130           0.59     ]
    1 : [  530           1.33     ]
    2 : [  960           2.03     ]
    3 : [ 1400           2.72     ]
    4 : [ 1830           2.83     ]
    >>> print("Inputs:", fm.model.getInputDescription())
    Inputs: [Q, Ks, Zv, Zm, B, L, Zb, Hd]
    >>> print("Output:", fm.model.getOutputDescription())
    Output: [H, S, C]

    Get the height model.

    >>> heightInputDistribution, heightModel = fm.getHeightModel()
    >>> print("Inputs:", heightModel.getInputDescription())
    Inputs: [Q,Ks,Zv,Zm]
    >>> print("Outputs:", heightModel.getOutputDescription())
    Outputs: [H]

    Get the flooding model with high Hd scenario.

    >>> fm = flood_model.FloodModel(distributionHdLow=False)
    """

    def __init__(self, trueKs=30.0, trueZv=50.0, trueZm=55.0, distributionHdLow=True):
        self.trueKs = trueKs
        self.trueZv = trueZv
        self.trueZm = trueZm
        self.distributionHdLow = distributionHdLow

        # Q
        self.Q = ot.TruncatedDistribution(
            ot.Gumbel(558.0, 1013.0), 0.0, ot.TruncatedDistribution.LOWER
        )
        self.Q.setName("Q")
        self.Q.setDescription(["Q (m3/s)"])

        # Ks
        self.Ks = ot.TruncatedDistribution(
            ot.Normal(30.0, 7.5), 0.0, ot.TruncatedDistribution.LOWER
        )
        self.Ks.setName("Ks")
        self.Ks.setDescription(["Ks"])

        # Zv
        self.Zv = ot.Uniform(49.0, 51.0)
        self.Zv.setName("Zv")
        self.Zv.setDescription(["Zv (m)"])

        # Zm
        self.Zm = ot.Uniform(54.0, 56.0)
        self.Zm.setName("Zm")
        self.Zm.setDescription(["Zm (m)"])

        # B
        self.B = ot.Triangular(295.0, 300.0, 305.0)
        self.B.setName("B")
        self.B.setDescription(["B (m)"])

        # L
        self.L = ot.Triangular(4990.0, 5000.0, 5010.0)
        self.L.setName("L")
        self.L.setDescription(["L (m)"])

        # Zb
        self.Zb = ot.Triangular(55.0, 55.5, 56.0)
        self.Zb.setName("Zb")
        self.Zb.setDescription(["Zb (m)"])

        # Hd
        if distributionHdLow:
            self.Hd = ot.Uniform(2.0, 4.0)
        else:
            self.Hd = ot.Uniform(7.0, 9.0)
        self.Hd.setName("Hd")
        self.Hd.setDescription(["Hd (m)"])

        # Formula
        formula = "var alpha := (Zm - Zv) / L;"
        formula += "H := (Q / (Ks * B * sqrt(alpha)))^(3.0 / 5.0);"
        formula += "var Zc := H + Zv;"
        formula += "var Zd := Zb + Hd;"
        formula += "S := Zc - Zd;"
        formula += (
            "var Cost_Flooding := if (S < 0, 0.2 - 0.8 * expm1(-1000 / S^4), 1.0);"
        )
        formula += "var Cost_Dyke := max(Hd, 8.0) / 20.0;"
        formula += "C := Cost_Flooding + Cost_Dyke;"

        self.model = ot.SymbolicFunction(
            ["Q", "Ks", "Zv", "Zm", "B", "L", "Zb", "Hd"], ["H", "S", "C"], formula
        )

        self.distribution = ot.JointDistribution(
            [self.Q, self.Ks, self.Zv, self.Zm, self.B, self.L, self.Zb, self.Hd]
        )
        self.dim = self.distribution.getDimension()

        self.data = ot.Sample(
            [
                [130.0, 0.59],
                [530.0, 1.33],
                [960.0, 2.03],
                [1400.0, 2.72],
                [1830.0, 2.83],
                [2260.0, 3.50],
                [2700.0, 3.82],
                [3130.0, 4.36],
                [3560.0, 4.63],
                [4010.0, 4.96],
            ]
        )
        self.data.setDescription(["Q ($m^3/s$)", "H (m)"])

    def getHeightModel(
        self,
        L=5000.0,
        B=300.0,
        Zb=55.5,
        Hd=3.0,
    ):
        """
        Return the height model with corresponding input distribution

        Parameters
        ----------
        L : float, optional
            The value of the river length. The default is 5000.0.

        B : float, optional
            The value of the river width. The default is 300.0.

        Zb : float, optional
            The level (altitude) of the bank. The default is 55.5.

        Hd : float, optional
            The height of the dyke. The default is 3.0.

        Returns
        -------
        heightInputDistribution : :class:`~openturns.Distribution`
            The joint input distribution of (Q, Ks, Zv, Zm).
        heightModel : :class:`~openturns.Function`
            The function with (Q, Ks, Zv, Zm) as input and (H) as output.
        """
        parametricModel = ot.ParametricFunction(
            self.model,
            [4, 5, 6, 7],
            [B, L, Zb, Hd],
        )
        heightModel = parametricModel.getMarginal(0)
        heightInputDistribution = self.distribution.getMarginal(range(4))
        return heightInputDistribution, heightModel



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


distribution = ot.Burr(2.5, 1.5)
size = 10000
sample = distribution.getSample(size)
factory = ot.BurrFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedBurr = factory.buildAsBurr(sample)
print("Burr          =", distribution)
print("Estimated burr=", estimatedBurr)
estimatedBurr = factory.buildAsBurr()
print("Default burr=", estimatedBurr)
estimatedBurr = factory.buildAsBurr(distribution.getParameter())
print("Burr from parameters=", estimatedBurr)

try:
    estimatedBurr = factory.build(ot.Normal(1e-3, 1e-5).getSample(100))
    print("Estimated burr=", estimatedBurr)
except Exception:
    pass

ot.RandomGenerator.SetSeed(0)
try:
    estimatedBurr = factory.build(
        ot.FiniteDiscreteDistribution(
            ot.LogNormal(7.71, 1.0056).getSample(500)
        ).getSupport()
    )
    # print('Estimated burr=', estimatedBurr)
except Exception:
    pass



"""
Create a parametric spectral density function
=============================================
"""

# %%
# This example illustrates how the User can create a density spectral
# function from parametric models.
#
# The library implements the *Cauchy spectral model* as a parametric model for the spectral density function
# :math:`S`.
#
# The library defines this model thanks to the object :class:`~openturns.CauchyModel`.

# %%
import openturns as ot


# %%
# 1. Define a spectral density function from correlation matrix
amplitude = [1.0, 2.0, 3.0]
scale = [4.0, 5.0, 6.0]
spatialCorrelation = ot.CorrelationMatrix(3)
spatialCorrelation[0, 1] = 0.8
spatialCorrelation[0, 2] = 0.6
spatialCorrelation[1, 2] = 0.1
spectralModel_Corr = ot.CauchyModel(amplitude, scale, spatialCorrelation)
spectralModel_Corr

# %%
# 2. Define a spectral density function from a covariance matrix
spatialCovariance = ot.CovarianceMatrix(3)
spatialCovariance[0, 0] = 4.0
spatialCovariance[1, 1] = 5.0
spatialCovariance[2, 2] = 6.0
spatialCovariance[0, 1] = 1.2
spatialCovariance[0, 2] = 0.9
spatialCovariance[1, 2] = -0.2
spectralModel_Cov = ot.CauchyModel(scale, spatialCovariance)
spectralModel_Cov



#! /usr/bin/env python

import openturns as ot

# First, build two functions from R^3->R^2
functions = list()
functions.append(
    ot.SymbolicFunction(
        ["x1", "x2", "x3"],
        [
            "x1^3 * sin(x2 + 2.5 * x3) - (x1 + x2)^2 / (1.0 + x3^2)",
            "x1^1 * sin(x3 + 2.5 * x1) - (x2 + x3)^2 / (1.0 + x1^2)",
        ],
    )
)
functions.append(
    ot.SymbolicFunction(
        ["x1", "x2", "x3"],
        [
            "exp(-x1 * x2 + x3) / cos(1.0 + x2 * x3 - x1)",
            "exp(-x2 * x3 + x1) / cos(1.0 + x3 * x1 - x2)",
        ],
    )
)
# Second, build the function
myFunction = ot.AggregatedFunction(functions)
inPoint = ot.Point([1.2, 2.3, 3.4])
print("myFunction=", myFunction)
print("Value at ", inPoint, "=", myFunction(inPoint))
print("Gradient at ", inPoint, "=", myFunction.gradient(inPoint))
ot.PlatformInfo.SetNumericalPrecision(5)
print("Hessian at ", inPoint, "=", myFunction.hessian(inPoint))
for i in range(myFunction.getOutputDimension()):
    print("Marginal ", i, "=", myFunction.getMarginal(i))
print("Marginal (0,1)=", myFunction.getMarginal([0, 1]))
print("Marginal (0,2)=", myFunction.getMarginal([0, 2]))
print("Marginal (1,2)=", myFunction.getMarginal([1, 2]))



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

krawtchouk = ot.KrawtchoukFactory(10, 0.3)
for i in range(10):
    print("krawtchouk(", i, ")=", krawtchouk.build(i))
roots = krawtchouk.getRoots(10)
print("krawtchouk(10) roots=", repr(roots))
nodes, weights = krawtchouk.getNodesAndWeights(10)
print("krawtchouk(10) nodes=", nodes, "and weights=", weights)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Instantiate one distribution object
dim = 3
R = ot.CorrelationMatrix(dim)
for i in range(dim - 1):
    R[i, i + 1] = 0.25
copula = ot.SklarCopula(ot.Distribution(ot.Normal([1.0, 2.0, 3.0], [2.0, 3.0, 1.0], R)))
copulaRef = ot.NormalCopula(R)
print("Copula ", repr(copula))
print("Copula ", copula)
print("Mean      =", repr(copula.getMean()))
print("Mean (ref)=", repr(copulaRef.getMean()))
ot.ResourceMap.SetAsUnsignedInteger("GaussKronrod-MaximumSubIntervals", 20)
ot.ResourceMap.SetAsScalar("GaussKronrod-MaximumError", 1.0e-4)
print("Covariance      =", repr(copula.getCovariance()))
ot.ResourceMap.SetAsUnsignedInteger("GaussKronrod-MaximumSubIntervals", 100)
ot.ResourceMap.SetAsScalar("GaussKronrod-MaximumError", 1.0e-12)
print("Covariance (ref)=", repr(copulaRef.getCovariance()))

# Is this copula an elliptical distribution?
print("Elliptical distribution= ", copula.isElliptical())

# Is this copula elliptical ?
print("Elliptical copula= ", copula.hasEllipticalCopula())

# Is this copula independent ?
print("Independent copula= ", copula.hasIndependentCopula())

# Test for realization of distribution
oneRealization = copula.getRealization()
print("oneRealization=", repr(oneRealization))

# Test for sampling
size = 10
oneSample = copula.getSample(size)
print("oneSample=", repr(oneSample))

# Test for sampling
size = 1000
anotherSample = copula.getSample(size)
print("anotherSample mean=", repr(anotherSample.computeMean()))
print("anotherSample covariance=", repr(anotherSample.computeCovariance()))

# Define a point
point = ot.Point(dim, 0.2)

# Show PDF and CDF of point
pointPDF = copula.computePDF(point)
pointCDF = copula.computeCDF(point)
pointPDFRef = copulaRef.computePDF(point)
pointCDFRef = copulaRef.computeCDF(point)
print(
    "Point = ", repr(point), " pdf      =%.6f" % pointPDF, " cdf      =%.6f" % pointCDF
)
print(
    "Point = ",
    repr(point),
    " pdf (ref)=%.6f" % pointPDFRef,
    " cdf (ref)=%.6f" % pointCDFRef,
)

# Get 50% quantile
quantile = copula.computeQuantile(0.5)
quantileRef = copulaRef.computeQuantile(0.5)
print("Quantile      =", repr(quantile))
print("Quantile (ref)=", repr(quantileRef))
print("CDF(quantile)=%.6f" % copula.computeCDF(quantile))
# Get 95% survival function
inverseSurvival = ot.Point(copula.computeInverseSurvivalFunction(0.95))
print("InverseSurvival=", repr(inverseSurvival))
print(
    "Survival(inverseSurvival)=%.6f" % copula.computeSurvivalFunction(inverseSurvival)
)
# Takes too much time
# print("entropy=%.6f" % copula.computeEntropy())
# Confidence regions
if copula.getDimension() <= 2:
    threshold = ot.Point()
    print(
        "Minimum volume interval=", copula.computeMinimumVolumeInterval(0.95, threshold)
    )
    print("threshold=", threshold)
    beta = ot.Point()
    levelSet = copula.computeMinimumVolumeLevelSet(0.95, beta)
    print("Minimum volume level set=", levelSet)
    print("beta=", beta)
    print(
        "Bilateral confidence interval=",
        copula.computeBilateralConfidenceInterval(0.95, beta),
    )
    print("beta=", beta)
    print(
        "Unilateral confidence interval (lower tail)=",
        copula.computeUnilateralConfidenceInterval(0.95, False, beta),
    )
    print("beta=", beta)
    print(
        "Unilateral confidence interval (upper tail)=",
        copula.computeUnilateralConfidenceInterval(0.95, True, beta),
    )
    print("beta=", beta)

# Extract the marginals
for i in range(dim):
    margin = copula.getMarginal(i)
    marginRef = copulaRef.getMarginal(i)
    print("margin=", repr(margin))
    print("margin PDF      =%.6f" % margin.computePDF(ot.Point(1, 0.25)))
    print("margin PDF (ref)=%.6f" % marginRef.computePDF(ot.Point(1, 0.25)))
    print("margin CDF      =%.6f" % margin.computeCDF(ot.Point(1, 0.25)))
    print("margin CDF (ref)=%.6f" % marginRef.computeCDF(ot.Point(1, 0.25)))
    print("margin quantile      =", repr(margin.computeQuantile(0.95)))
    print("margin quantile (ref)=", repr(marginRef.computeQuantile(0.95)))
    print("margin realization=", repr(margin.getRealization()))

# Extract a 2-D marginal
indices = ot.Indices([1, 0])
print("indices=", repr(indices))
margin = copula.getMarginal(indices)
marginRef = copulaRef.getMarginal(indices)
print("margin=", repr(margin))
print("margin PDF      =%.6f" % margin.computePDF(ot.Point(2, 0.25)))
print("margin PDF (ref)=%.6f" % marginRef.computePDF(ot.Point(2, 0.25)))
print("margin CDF      =%.6f" % margin.computeCDF(ot.Point(2, 0.25)))
print("margin CDF (ref)=%.6f" % marginRef.computeCDF(ot.Point(2, 0.25)))
print("margin quantile      =", repr(margin.computeQuantile(0.95)))
print("margin quantile (ref)=", repr(marginRef.computeQuantile(0.95)))
print("margin realization=", repr(margin.getRealization()))

ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(copula)
validation.run()

# tbb nested parallelism issue
student = ot.Student(3.0, [1.0] * 2, [3.0] * 2, ot.CorrelationMatrix(2))
copula = ot.SklarCopula(student)
pdf_graph = copula.drawPDF()
cdf_graph = copula.drawCDF()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Realization issued from a SpectralProcess
dimension = 1

# Parameters of the distribution
N = 51
t0 = 0.0
dt = 0.1
myTimeGrid = ot.RegularGrid(t0, dt, N)

# Create a Sample
# parameters of gaussien impose a few risk to get negative values
mySample = ot.Uniform().getSample(N)

# get a realization from distribution
myRealization = ot.TimeSeries(myTimeGrid, mySample)

# Create the lambda parameter
lambdaVector = ot.Point(dimension)
for index in range(dimension):
    lambdaVector[index] = (index + 2.0) * 0.1

myInverseBoxCox = ot.InverseBoxCoxTransform(lambdaVector)

print("myInverseBoxCox=", myInverseBoxCox)

# Get the input and output dimension
print("myInverseBoxCox input dimension = ", myInverseBoxCox.getInputDimension())
print("myInverseBoxCox output dimension = ", myInverseBoxCox.getOutputDimension())

# Evaluation of the InverseBoxCoxTransform on the realization
print("input time series =")
print(myRealization)
print("output time series = ")
print(myInverseBoxCox(myRealization))

print("gradient=", myInverseBoxCox.gradient([0.5]))
print("hessian=", myInverseBoxCox.hessian([0.5]))

# Call the getInverse method
myBoxCox = myInverseBoxCox.getInverse()
print("myBoxCox = ", myBoxCox)

# Get the number of calls
print("number of call(s) : ", myInverseBoxCox.getCallsNumber())



#! /usr/bin/env python

# %%
import openturns as ot
import openturns.testing as ott
import os

# %%
ot.TESTPREAMBLE()

# Instantiate one distribution object
distribution = ot.Binomial(15, 0.7)
print("Distribution ", repr(distribution))
print("Distribution ", distribution)

# Is this distribution elliptical ?
print("Elliptical = ", distribution.isElliptical())

# Is this distribution continuous ?
print("Continuous = ", distribution.isContinuous())

# Test for realization of distribution
oneRealization = distribution.getRealization()
print("oneRealization=", repr(oneRealization))

# Define a point
point = ot.Point(distribution.getDimension(), 5.0)
print("Point= ", repr(point))

# PDF value
PDF = distribution.computePDF(point)
print("pdf     =%.6f" % PDF)

# derivative of the PDF with regards the parameters of the distribution
CDF = distribution.computeCDF(point)
print("cdf=%.6f" % CDF)
# quantile
quantile = distribution.computeQuantile(0.95)
print("quantile=", repr(quantile))
print("cdf(quantile)=%.6f" % distribution.computeCDF(quantile))
print("entropy=%.6f" % distribution.computeEntropy())
mean = distribution.getMean()
print("mean=", repr(mean))
standardDeviation = distribution.getStandardDeviation()
print("standard deviation=", repr(standardDeviation))
skewness = distribution.getSkewness()
print("skewness=", repr(skewness))
kurtosis = distribution.getKurtosis()
print("kurtosis=", repr(kurtosis))
covariance = distribution.getCovariance()
print("covariance=", repr(covariance))
parameters = distribution.getParametersCollection()
print("parameters=", repr(parameters))
print("Standard representative=", distribution.getStandardRepresentative())
# Confidence interval
alpha = 0.05
bounds = distribution.computeBilateralConfidenceInterval(1 - alpha)
print("%.2f%% bilateral confidence interval" % ((1 - alpha) * 100), " =", bounds)

# %%
# check survival at upper bound
distribution = ot.Binomial(10, 1.0)
assert distribution.computeCDF(10.0) == 1.0
assert distribution.computeComplementaryCDF(10.0) == 0.0
assert distribution.computeSurvivalFunction(10.0) == 0.0

# negative quantile bug
distribution = ot.Binomial(3, 0.5)
assert distribution.computeScalarQuantile(0.9, True) == 0

# %%
print("Check on dataset")
separator = ","
path = os.path.join(os.path.dirname(__file__), "t_Binomial_std.csv")
sample = ot.Sample.ImportFromTextFile(path, separator)

rtol = 1.0e-12
atol = ot.SpecFunc.MinScalar
sample_size = sample.getSize()
for i in range(sample_size):
    x, n, pr, expected_pdf, expected_cdfp, expected_cdfq = sample[i]
    x = int(x)
    n = int(n)
    distribution = ot.Binomial(n, pr)
    computed_p = distribution.computePDF(x)
    computed_cdf = distribution.computeCDF(x)
    computed_ccdf = distribution.computeComplementaryCDF(x)
    computed_surv = distribution.computeSurvivalFunction(x)
    print(
        f"row = {i}/{sample_size}, x = {x}, n = {n}, pr = {pr}, "
        f"P(X = x) = {expected_pdf}, CDF = {expected_cdfp}, Compl. CDF = {expected_cdfq}"
    )
    # Check PDF
    print(
        f"    computePDF. Computed = {computed_p:.11e}, expected = {expected_pdf:.11e}"
    )
    ott.assert_almost_equal(computed_p, expected_pdf, rtol, atol)
    # Check CDF
    print(
        f"    computeCDF. Computed = {computed_cdf:.11e}, expected = {expected_cdfp:.11e}"
    )
    ott.assert_almost_equal(computed_cdf, expected_cdfp, rtol, atol)
    # Check complementary CDF
    print(
        f"    computeComplementaryCDF. Computed = {computed_ccdf:.11e}, "
        f"expected = {expected_cdfq:.11e}"
    )
    ott.assert_almost_equal(computed_ccdf, expected_cdfq, rtol, atol)
    # Check Survival
    print(
        f"    computeSurvivalFunction. Computed = {computed_surv:.11e}, "
        f"expected = {computed_ccdf:.11e}"
    )
    ott.assert_almost_equal(computed_surv, computed_ccdf, rtol, atol)
    # Check quantile
    if pr == 0.0 and x < n:
        # The function is not invertible
        # for this particular input.
        continue
    elif expected_cdfp < expected_cdfq:
        computed_x = int(distribution.computeQuantile(computed_cdf)[0])
        cdfXM1 = distribution.computeCDF(computed_x - 1)
        cdfX = distribution.computeCDF(computed_x)
        print(
            f"    computeQuantile (A). Computed X = {computed_x}, "
            f"F(X - 1) = {cdfXM1:.7e}, F(X) = {cdfX:.7e}"
        )
        if computed_cdf == 0.0:
            assert computed_x == 0.0
        else:
            assert cdfXM1 < computed_cdf and computed_cdf <= cdfX
    else:
        computed_x = int(distribution.computeQuantile(computed_ccdf, True)[0])
        print(f"    computeQuantile (B). Computed X = {computed_x}, expected = {x}")
        assert x == computed_x
    if i > 0:
        ott.assert_almost_equal(
            distribution.computeScalarQuantile(expected_cdfp), x, 0.0, 1.0
        )  # Can be off by 1 unit


# %%
# quantile bug
alpha = 0.05
beta = 0.05
for n in range(59, 100):
    d = ot.Binomial(n, alpha)
    k = d.computeQuantile(beta)[0]
    p1 = d.computeCDF(k - 1)
    p2 = d.computeCDF(k)
    ok = p1 < beta <= p2
    print(f"n={n} k={k:.0f} p(k-1)={p1:.4f} p(k)={p2:.4f} ok={ok}")
    assert ok

ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(distribution)
validation.run()

# Corner cases

# 2147483647 is the maximum integer.
# Values greater than this are not doubles anymore.
binomial = ot.Binomial()
N = 2147483647
binomial.setN(N)
binomial.setP(1.0 / N)
computed = binomial.computePDF(1.0)
expected = 0.3678794
ott.assert_almost_equal(computed, expected, 1.0e-6, 0.0)

computed = binomial.computePDF(2.0)
expected = 0.1839397
ott.assert_almost_equal(computed, expected, 1.0e-6, 0.0)

# Extreme inputs
binomial = ot.Binomial()
binomial.setN(9999)
binomial.setP(0.5)
computed = binomial.computePDF(4999.0)
expected = 0.0079786461393821558191
ott.assert_almost_equal(computed, expected, 1.0e-7, 0.0)

# Check pdf for values of P closer to 1
binomial = ot.Binomial()
binomial.setN(2)
binomial.setP(0.9999)
computed = binomial.computePDF(1.0)
expected = 1.999799999999779835e-04
ott.assert_almost_equal(computed, expected, 1e-12, 0.0)



"""
Create a linear combination of functions
========================================
"""

# %%
# In this example we are going to build a linear combination of several functions.
#
# .. math::
#    f = \sum \alpha_i f_i
#
# Here
#
# .. math::
#    f_1(x) = [x_1^2+x_2,x_1+x_2+x_3]
#    f_2(x) = [x_1 + 2 * x_2 + x_3,x_1+x_2+x_3]
#
# with coefficients
#
# .. math::
#    a_1 = 2, a_2 = 4
#

# %%
import openturns as ot


# %%
# assume a list of functions to combine
functions = list()
functions.append(ot.SymbolicFunction(["x1", "x2", "x3"], ["x1^2 + x2", "x1 + x2 + x3"]))
functions.append(
    ot.SymbolicFunction(["x1", "x2", "x3"], ["x1 + 2 * x2 + x3", "x1 + x2 - x3"])
)

# %%
# create the combination function
coefficients = [2.0, 4.0]
function = ot.LinearCombinationFunction(functions, coefficients)

# %%
# evaluate the function
x = [1.0, 2.0, 3.0]
y = function(x)
print("x=", x, "y=", y)



"""
Estimate quantile confidence intervals
======================================
"""

# %%
# In this example, we introduce two methods to estimate a confidence interval of the
# :math:`\alpha` level quantile (:math:`\alpha \in [0,1]`) of the distribution of
# a random
# variable :math:`X` of dimension 1. Both methods use the order statistics to estimate:
#
# - an asymptotic confidence interval with confidence level :math:`\beta \in [0,1]`,
# - an exact upper bounded confidence interval with confidence level :math:`\beta \in [0,1]`.
#
# In this example, we consider the quantile of level :math:`\alpha = 95\%`,
# with a confidence level of :math:`\beta = 90\%`.
#
# See  :ref:`quantile_confidence_estimation`  and :ref:`quantile_asymptotic_confidence_estimation` to get details on the signification of these confidence interval.
import openturns as ot
import openturns.experimental as otexp
import math as m


# %%
# We consider a random vector which is the output of a model and an input distribution.
model = ot.SymbolicFunction(["x1", "x2"], ["x1^2+x2"])
R = ot.CorrelationMatrix(2)
R[0, 1] = -0.6
inputDist = ot.Normal([0.0, 0.0], R)
inputDist.setDescription(["X1", "X2"])
inputVector = ot.RandomVector(inputDist)

# %%
# Create the output random vector
output = ot.CompositeRandomVector(model, inputVector)

# %%
# We define the level :math:`\alpha` of the quantile and the confidence level :math:`\beta`.
alpha = 0.95
beta = 0.90

# %%
# We generate a sample of the variable.
n = 10**4
sample = output.getSample(n)

# %%
# We get the empirical estimator of the :math:`\alpha` level quantile which is the
# :math:`\lfloor \sampleSize \alpha \rfloor` -th order statistics evaluated on
# the sample.
empiricalQuantile = sample.computeQuantile(alpha)
print(empiricalQuantile)

# %%
# The asymptotic confidence interval of level :math:`\beta` is :math:`\left[ X_{(i_n)}, X_{(j_n)}\right]`
# such that:
#
# .. math::
#
#     i_\sampleSize & = \left\lfloor \sampleSize \alpha - \sqrt{\sampleSize} \; z_{\frac{1+\beta}{2}} \; \sqrt{\alpha(1 - \alpha)} \right\rfloor\\
#     j_\sampleSize & = \left\lfloor \sampleSize \alpha + \sqrt{\sampleSize} \; z_{\frac{1+\beta}{2}} \;  \sqrt{\alpha(1 - \alpha)} \right\rfloor
#
# where  :math:`z_{\frac{1+\beta}{2}}` is the :math:`\frac{1+\beta}{2}` level quantile of the standard normal distribution (see [delmas2006]_ proposition 11.1.13).
#
# Then we have:
#
# .. math::
#
#     \lim\limits_{\sampleSize \rightarrow +\infty} \Prob{x_{\alpha} \in \left[ X_{(i_\sampleSize,\sampleSize)}, X_{(j_\sampleSize,\sampleSize)}\right]} = \beta
#
a_beta = ot.Normal(1).computeQuantile((1.0 + beta) / 2.0)[0]
i_n = int(n * alpha - a_beta * m.sqrt(n * alpha * (1.0 - alpha)))
j_n = int(n * alpha + a_beta * m.sqrt(n * alpha * (1.0 - alpha)))
print(i_n, j_n)

# %%
# Get the sorted sample
sortedSample = sample.sort()

# %%
# Get the asymptotic confidence interval :math:`\left[ X_{(i_n)}, X_{(j_n)}\right]`.
#
# Care: the index in the sorted sample is :math:`i_n-1` and :math:`j_n-1`
# because python numbering starts at 0.
infQuantile = sortedSample[i_n - 1]
supQuantile = sortedSample[j_n - 1]
print(infQuantile, empiricalQuantile, supQuantile)

# %%
# This can be done using :class:`~openturns.experimental.QuantileConfidence`
# when the ranks :math:`i_n` and :math:`j_n` are directly given in :math:`\llbracket 0, \sampleSize-1 \rrbracket`.
algo = otexp.QuantileConfidence(alpha, beta)
i_n, j_n = algo.computeAsymptoticBilateralRank(n)
ci = algo.computeAsymptoticBilateralConfidenceInterval(sample)
print(f"asymptotic ranks={[i_n, j_n]} CI={ci}")

# %%
# The empirical quantile was estimated with the :math:`\lfloor \sampleSize\alpha \rfloor` -th order statistics evaluated on
# the sample of size :math:`\sampleSize`.
# We define :math:`i = \sampleSize-\lfloor \sampleSize\alpha \rfloor` and we evaluate the minimum sample size :math:`\tilde{\sampleSize}` that
# ensures that the :math:`(\tilde{\sampleSize}-i)` order statistics is greater than :math:`x_{\alpha}` with the confidence :math:`\beta`.
i = n - int(n * alpha)
minSampleSize = algo.computeUnilateralMinimumSampleSize(i, True)
print(minSampleSize)

# %%
# Now we can evaluate the upper bounded confidence interval:
# we generate a sample with the previous minimum size and extract the empirical quantile of order :math:`(\tilde{\sampleSize}-i)`.
sample = output.getSample(minSampleSize)
upperBoundQuantile = sample.sort()[-i - 1]
print(upperBoundQuantile)

# %%
# We can also find this rank back from the size.
k = algo.computeUnilateralRank(minSampleSize, True)
print(k, minSampleSize - i - 1)

# %%
# The quantile bound is given in the same manner from the sample and the rank.
ci = algo.computeUnilateralConfidenceInterval(sample, True)
print(ci)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# First, build two functions from R^3->R
inVar = ot.Description(3)
inVar[0] = "x1"
inVar[1] = "x2"
inVar[2] = "x3"
formula = ot.Description(1)
formula[0] = "x1^3 * sin(x2 + 2.5 * x3) - (x1 + x2)^2 / (1.0 + x3^2)"
functions = []
functions.append(ot.SymbolicFunction(inVar, formula))
formula[0] = "exp(-x1 * x2 + x3) / cos(1.0 + x2 * x3 - x1)"
functions.append(ot.SymbolicFunction(inVar, formula))
# Second, build the weights
coefficients = ot.Point(2)
coefficients[0] = 0.3
coefficients[1] = 2.9
myGradient = ot.LinearCombinationGradient(
    ot.LinearCombinationEvaluation(functions, coefficients)
)
inPoint = ot.Point(3)
inPoint[0] = 1.2
inPoint[1] = 2.3
inPoint[2] = 3.4
print("myGradient=", myGradient)
print("Value at ", inPoint, "=", myGradient.gradient(inPoint))



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

sample = ot.Sample(3, 2)
sample[0] = ot.Point((1.0, 1.5))
sample[1] = ot.Point((2.0, 2.5))
sample[2] = ot.Point((3.0, 3.5))

factory = ot.FiniteDiscreteDistributionFactory()
estimatedDistribution = factory.build(sample)
print("sample=", repr(sample))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedUserDefined = factory.buildAsFiniteDiscreteDistribution(sample)
print("Sample                =", sample)
print("Estimated FiniteDiscreteDistribution=", estimatedUserDefined)
estimatedUserDefined = factory.buildAsFiniteDiscreteDistribution()
print("Default FiniteDiscreteDistribution=", estimatedUserDefined)
estimatedDistribution = factory.build([[10]] * 2000)
print("Estimated distribution=", repr(estimatedDistribution))



"""
Compare covariance models
=========================
"""

# %%
# The main goal of this example is to briefly review the most important covariance models and compare them in terms of regularity of the trajectories.
#
# We first show how to define a covariance model, a temporal grid and a Gaussian process.
# We first consider the squared exponential covariance model and show how the trajectories are sensitive to its parameters.
# We show how to define a trend. In the final section, we compare the trajectories from exponential and Matérn covariance models.

# %%
# References
# ----------
#
# * Carl Edward Rasmussen and Christopher K. I. Williams (2006) Gaussian Processes for Machine Learning. Chapter 4: "Covariance Functions", www.GaussianProcess.org/gpml

# %%
# The anisotropic squared exponential model
# -----------------------------------------
#
# The :class:`~openturns.SquaredExponential` class allows one to define covariance models:
#
# * :math:`\sigma\in\mathbb{R}` is the amplitude parameter,
# * :math:`\boldsymbol{\theta}\in\mathbb{R}^d` is the scale.

# %%
import openturns.viewer as otv
import openturns as ot
import matplotlib.pyplot as plt

# %%
# Amplitude values
amplitude = [3.5]
# Scale values
scale = [1.5]
# Covariance model
myModel = ot.SquaredExponential(scale, amplitude)

# %%
# Gaussian processes
# ------------------
#
# In order to create a :class:`~openturns.GaussianProcess`, we must have:
#
# * a covariance model,
# * a grid.
#
# Optionnally, we can define a trend (we will see that later in the example). By default, the trend is zero.
#
# We consider the domain :math:`\mathcal{D}=[0,10]`. We discretize this domain with 100 cells (which corresponds to 101 nodes), with steps equal to 0.1 starting from 0:
#
# .. math::
#    (x_0=x_{min}=0,\:x_1=0.1,\:\ldots,\:x_n=x_{max}=10).
#

# %%
xmin = 0.0
step = 0.1
n = 100
myTimeGrid = ot.RegularGrid(xmin, step, n + 1)
graph = myTimeGrid.draw()
graph.setTitle("Regular grid")
view = otv.View(graph)

# %%
# Then we create the Gaussian process (by default the trend is zero).

# %%
process = ot.GaussianProcess(myModel, myTimeGrid)

# %%
# Then we generate 10 trajectores with the `getSample` method. This trajectories are in a :class:`~openturns.ProcessSample`.

# %%
nbTrajectories = 10
sample = process.getSample(nbTrajectories)
type(sample)

# %%
# We can draw the trajectories with `drawMarginal`.

# %%
graph = sample.drawMarginal(0)
graph.setTitle("amplitude=%.3f, scale=%.3f" % (amplitude[0], scale[0]))
view = otv.View(graph)


# %%
# In order to make the next examples easier, we define a function which plots a given number of trajectories from a Gaussian process based on a covariance model.


# %%
def plotCovarianceModel(myCovarianceModel, myTimeGrid, nbTrajectories):
    """Plots the given number of trajectories with given covariance model."""
    process = ot.GaussianProcess(myCovarianceModel, myTimeGrid)
    sample = process.getSample(nbTrajectories)
    graph = sample.drawMarginal(0)
    graph.setTitle("")
    return graph


# %%
# The amplitude parameter sets the variance of the process. A greater amplitude increases the chances of getting larger absolute values of the process.

# %%
amplitude = [7.0]
scale = [1.5]
myModel = ot.SquaredExponential(scale, amplitude)
graph = plotCovarianceModel(myModel, myTimeGrid, 10)
graph.setTitle("amplitude=%.3f, scale=%.3f" % (amplitude[0], scale[0]))
view = otv.View(graph)

# %%
# Modifying the scale parameter is here equivalent to stretch or contract the "time" :math:`x`.

# %%
amplitude = [3.5]
scale = [0.5]
myModel = ot.SquaredExponential(scale, amplitude)
graph = plotCovarianceModel(myModel, myTimeGrid, 10)
graph.setTitle("amplitude=%.3f, scale=%.3f" % (amplitude[0], scale[0]))
view = otv.View(graph)

# %%
# Define the trend
# ----------------
#
# The trend is a deterministic function. With the :class:`~openturns.GaussianProcess` class, the associated process is the sum of a trend and a Gaussian process with zero mean.

# %%
f = ot.SymbolicFunction(["x"], ["2*x"])
fTrend = ot.TrendTransform(f, myTimeGrid)

# %%
amplitude = [3.5]
scale = [1.5]
myModel = ot.SquaredExponential(scale, amplitude)
process = ot.GaussianProcess(fTrend, myModel, myTimeGrid)

# %%
# sphinx_gallery_thumbnail_number = 5
nbTrajectories = 10
sample = process.getSample(nbTrajectories)
graph = sample.drawMarginal(0)
graph.setTitle("amplitude=%.3f, scale=%.3f" % (amplitude[0], scale[0]))
view = otv.View(graph)

# %%
# Other covariance models
# -----------------------
#
# There are other covariance models. The models which are used more often are the following:
#
# * :class:`~openturns.SquaredExponential`. The generated processes can be derivated in mean square at all orders.
# * :class:`~openturns.MaternModel`. When :math:`\nu\rightarrow+\infty`, it converges to the squared exponential model.
#   This model can be derivated :math:`k` times only if :math:`k<\nu`.
#   In other words, when :math:`\nu` increases, then the trajectories are more and more regular.
#   The particular case :math:`\nu=1/2` is the exponential model.
#   The most commonly used values are :math:`\nu=3/2` and :math:`\nu=5/2`, which produce trajectories that are, in terms of regularity,
#   in between the squared exponential and the exponential models.
# * :class:`~openturns.ExponentialModel`. The associated process is continuous, but not differentiable.
#

# %%
# The Matérn and exponential models
# ---------------------------------

# %%
amplitude = [1.0]
scale = [1.0]
nu1, nu2, nu3 = 2.5, 1.5, 0.5
myModel1 = ot.MaternModel(scale, amplitude, nu1)
myModel2 = ot.MaternModel(scale, amplitude, nu2)
myModel3 = ot.MaternModel(scale, amplitude, nu3)

# %%
nbTrajectories = 10
graph1 = plotCovarianceModel(myModel1, myTimeGrid, nbTrajectories)
graph2 = plotCovarianceModel(myModel2, myTimeGrid, nbTrajectories)
graph3 = plotCovarianceModel(myModel3, myTimeGrid, nbTrajectories)

# %%
fig = plt.figure(figsize=(20, 6))
ax1 = fig.add_subplot(1, 3, 1)
_ = otv.View(graph1, figure=fig, axes=[ax1])
_ = ax1.set_title("Matern 5/2")
ax2 = fig.add_subplot(1, 3, 2)
_ = otv.View(graph2, figure=fig, axes=[ax2])
_ = ax2.set_title("Matern 3/2")
ax3 = fig.add_subplot(1, 3, 3)
_ = otv.View(graph3, figure=fig, axes=[ax3])
_ = ax3.set_title("Matern 1/2")

# %%
# We see than, when :math:`\nu` increases, then the trajectories are smoother and smoother.

# %%
myExpModel = ot.ExponentialModel(scale, amplitude)

# %%
graph = plotCovarianceModel(myExpModel, myTimeGrid, nbTrajectories)
graph.setTitle("Exponential")
view = otv.View(graph)

# %%
# We see that the exponential model produces very irregular trajectories.
otv.View.ShowAll()



"""
Sample manipulation
===================
"""

# %%
# This example will describe the main statistical functionalities on data through the :class:`~openturns.Sample` object. The Sample is an output variable of interest.

# %%
import openturns as ot


# %%
# A typical example
# -----------------

# %%
# A recurring issue in uncertainty quantification is to perform analysis on an output variable of interest `Y` obtained through a model `f` and input parameters `X`.
# Here we shall consider the input parameters as two independent standard Normal distributions :math:`X=(X_1, X_2)`.
# We therefore use an :class:`~openturns.IndependentCopula` to describe the link between the two marginals.
#

# input parameters
inputDist = ot.JointDistribution([ot.Normal()] * 2, ot.IndependentCopula(2))
inputDist.setDescription(["X1", "X2"])

# %%
# We create a vector from the 2d-distribution created before :

# %%
inputVector = ot.RandomVector(inputDist)


# %%
# Suppose our model `f` is known and reads as :
#
# .. math::
#    f(x) = \begin{pmatrix}
#             x_1^2 + x_2 \\
#             x_1   + x_2^2
#           \end{pmatrix}
#
# We define our model `f` with a :class:`~openturns.SymbolicFunction`

# %%
f = ot.SymbolicFunction(["x1", "x2"], ["x1^2+x2", "x2^2+x1"])

# %%
# Our output vector is :math:`Y = f(X)`, the image of the inputVector by the model
outputVector = ot.CompositeRandomVector(f, inputVector)

# %%
# We can now get a sample out of `Y`, that is realizations (here 1000) of the random outputVector
size = 1000
sample = outputVector.getSample(size)

# %%
# The sample may be seen as a matrix of size :math:`1000 \times 2`. We print the 5 first samples (out of 1000) :

# %%
sample[:5]


# %%
# Basic operations on samples
# ---------------------------
# We have access to basic information about a sample such as
#
# - minimum and maximum per component
#
sample.getMin(), sample.getMax()

# %%
# - the range per component (max-min)
#
sample.computeRange()

# %%
# More elaborate functionalities are also available :
#

# %%
# - get the median per component
#
sample.computeMedian()

# %%
# - compute the covariance
#
sample.computeCovariance()

# %%
# - get the empirical 0.95 quantile per component
#
sample.computeQuantilePerComponent(0.95)

# %%
# - get the value of the empirical CDF at a point
#
point = [1.1, 2.2]
sample.computeEmpiricalCDF(point)


# %%
# Estimate the statistical moments
# --------------------------------

# %%
# Oftentimes, we need to estimate the first moments of the output data. We can then estimate statistical moments from the output sample :

# %%
# - estimate the moment of order 1 : mean
#
sample.computeMean()

# %%
# - estimate the standard deviation for each component
#
sample.computeStandardDeviation()

# %%
# - estimate the moment of order 2 : variance
#
sample.computeVariance()

# %%
# - estimate the moment of order 3 : skewness
#
sample.computeSkewness()

# %%
# - estimate the moment of order 4 : kurtosis
#
sample.computeKurtosis()

# %%
# Test the correlation
# --------------------

# %%
# Some statistical test for correlation are available :
#
# - get the sample linear correlation matrix :
#
sample.computeLinearCorrelation()

# %%
# - get  the sample Kendall correlation matrix :
#
sample.computeKendallTau()

# %%
#  - get  the sample Spearman correlation matrix :
#
sample.computeSpearmanCorrelation()



"""
Gaussian Process Regression: choose an arbitrary trend
======================================================
"""

# %%
# The goal of this example is to show how to configure an arbitrary trend in a Gaussian Process Regression metamodel
# (refer to :ref:`gaussian_process_regression`).
# In the :doc:`/auto_surrogate_modeling/gaussian_process_regression/plot_gpr_choose_trend`
# and :doc:`/auto_surrogate_modeling/gaussian_process_regression/plot_gpr_beam_trend` examples,
# we show how to configure a polynomial trend.
#
# In general, any collection of multivariate functions can be used as the
# `basis` argument of a :class:`~openturns.GaussianProcessFitter` or a
# :class:`~openturns.GaussianProcessRegression`.
# In practice, it might not be convenient to create a multivariate basis and
# this is why we sometimes create it by tensorization of univariate functions.
# In this example, we first use Legendre polynomials as our univariate functions,
# then we create an orthogonal polynomial basis corresponding to the input marginals.
#
# For this purpose, we use the :ref:`cantilever beam <use-case-cantilever-beam>` example.
from openturns.usecases import cantilever_beam
import openturns as ot


# %%
# Definition of the model
# -----------------------
#
# We load the cantilever beam use case
cb = cantilever_beam.CantileverBeam()

# %%
# We load the function (model) which evaluates the output `Y` depending on the inputs.
model = cb.model

# %%
# Then we define the distribution of the input random vector.
dimension = cb.dim  # number of inputs
input_dist = cb.distribution

# %%
# Create the design of experiments
# --------------------------------

# %%
# We consider a simple Monte-Carlo sampling as a design of experiments.
# This is why we generate an input sample using the `getSample` method of the distribution.
# Then we evaluate the output using the `model` function.
sampleSize_train = 20
X_train = input_dist.getSample(sampleSize_train)
Y_train = model(X_train)

# %%
# Create the Legendre basis
# -------------------------
#
# We first create a Legendre basis of univariate polynomials.
#
# The :class:`~openturns.LegendreFactory` class creates Legendre polynomials.
univariateFactory = ot.LegendreFactory()

# %%
# These polynomials are orthogonal with respect to the `Uniform` distribution on :math:`[-1, 1]`, as we can see.
univariateFactory.getMeasure()

# %%
# Even if the measure  `Uniform([-1, 1])` does not correspond to the input marginal distributions,
# these polynomials will, anyway, create a consistent trend for the Gaussian process regression metamodel.
#
# We use the same polynomial family for each input. The multivariate polynomials are created by tensorization of the univariate polynomials.
# The linear enumerate function specifies the order followed to create the multivariate polynomials.
polyColl = [univariateFactory] * dimension
enumerateFunction = ot.LinearEnumerateFunction(dimension)
productBasis = ot.OrthogonalProductPolynomialFactory(polyColl, enumerateFunction)

# %%
# The basis contains the first multivariate polynomials ordered according to the linear enumerate function.
functions = []
numberOfTrendCoefficients = 12
for i in range(numberOfTrendCoefficients):
    multivariatepolynomial = productBasis.build(i)
    print(multivariatepolynomial)
    functions.append(multivariatepolynomial)

# %%
basis = ot.Basis(functions)

# %%
# Create the Gaussian Process Regression metamodel
# ------------------------------------------------
#
# In order to create the  Gaussian process regression  metamodel, we choose the function basis created previously
# and a squared exponential covariance model.
covariance_model = ot.SquaredExponential([1.0] * dimension, [1.0])

# %%
# First, we estimate a Gaussian process approximating the model with the class
# :class:`~openturns.GaussianProcessFitter`.
algo_fit = ot.GaussianProcessFitter(X_train, Y_train, covariance_model, basis)
print("First run: algo GPFitter = ", algo_fit.getOptimizationAlgorithm())
algo_fit.setOptimizationAlgorithm(ot.TNC())
algo_fit.run()
fitter_result = algo_fit.getResult()

# %%
# Then, we condition the estimated gaussian process to
# make it interpolate the data set using the class
# :class:`~openturns.GaussianProcessRegression`.
gpr_algo = ot.GaussianProcessRegression(fitter_result)
gpr_algo.run()
gpr_result = gpr_algo.getResult()

# %%
# We get the Gaussian process regression metamodel.
gpr_metamodel = gpr_result.getMetaModel()

# %%
# The method :meth:`~openturns.GaussianProcessRegressionResult.getTrendCoefficients` returns the
# coefficients of the trend. We see that the number of coefficients in the trend corresponds to the number of
# functions in the basis.
print(gpr_result.getTrendCoefficients())

# %%
# We get the updated covariance model. The :class:`~openturns.SquaredExponential` model has one amplitude coefficient
# and 4 scale coefficients. This is because this covariance model is anisotropic : each of the 4 input variables
# is associated with its own scale coefficient.
print(gpr_result.getCovarianceModel())

# %%
# Create an orthogonal multivariate polynomial factory
# ----------------------------------------------------
#
# We suggest here to create a multivariate polynomial basis where each marginal polynomial family corresponds to the polynomial family orthogonal
# to the marginal input distribution. We use the class :class:`~openturns.OrthogonalProductPolynomialFactory` created from the input marginal distributions.
# This corresponds to the basis we usually use in the polynomial chaos expansion.
#
# We first create the mutlivariate polynomial basis as a tensorized product of the univariate polynomial basis orthogonal to the marginal input distributions.
multivariateBasis = ot.OrthogonalProductPolynomialFactory([cb.E, cb.F, cb.L, cb.II])

# %%
# Then we create the multivariate basis which has maximum degree equal to 2. There are 15 functions contained in the basis.
totalDegree = 2
enumerateFunction = multivariateBasis.getEnumerateFunction()
numberOfTrendCoefficients = enumerateFunction.getStrataCumulatedCardinal(totalDegree)
print("Number of functions in the basis: ", numberOfTrendCoefficients)

# %%
functions = []
for i in range(numberOfTrendCoefficients):
    multivariatepolynomial = productBasis.build(i)
    print(multivariatepolynomial)
    functions.append(multivariatepolynomial)

# %%
basis = ot.Basis(functions)

# %%
algo_fit = ot.GaussianProcessFitter(X_train, Y_train, covariance_model, basis)
print("Second run: algo GPFitter = ", algo_fit.getOptimizationAlgorithm())
algo_fit.setOptimizationAlgorithm(ot.TNC())
algo_fit.run()
fitter_result = algo_fit.getResult()
gpr_algo = ot.GaussianProcessRegression(fitter_result)
gpr_algo.run()
gpr_result = gpr_algo.getResult()
print(gpr_result.getTrendCoefficients())

# %%
# Conclusion
# ----------
#
# The trend that we have configured corresponds to the basis that we would have used in a full polynomial chaos expansioncomputed with least squares.
#
# Other extensions of this work would be:
#
# * to use a Fourier basis with :class:`~openturns.FourierSeriesFactory`,
# * wavelets with :class:`~openturns.HaarWaveletFactory`,
#
# or any other univariate factory.



"""
Gaussian Process Regression : cantilever beam model
===================================================
"""

# %%
# In this example, we create a Gaussian Process Regression (GPR) metamodel (refer to
# :ref:`gaussian_process_regression`) of the :ref:`cantilever beam <use-case-cantilever-beam>`.
# We use a squared exponential covariance kernel for the Gaussian process. In order to estimate the hyper-parameters, we use a design of experiments of size 20.
from openturns.usecases import cantilever_beam
import openturns as ot
import openturns.viewer as otv

# sphinx_gallery_thumbnail_number = 3

# %%
# Definition of the model
# -----------------------
#
# We load the cantilever beam use case :
cb = cantilever_beam.CantileverBeam()

# %%
# We define the function which evaluates the output depending on the inputs.
model = cb.model

# %%
# Then we define the distribution of the input random vector.
myDistribution = cb.distribution

# %%
# Create the design of experiments
# --------------------------------
#
# We consider a simple Monte Carlo sample as a design of experiments.
# This is why we generate an input sample using the method :meth:`~openturns.Distribution.getSample` of the
# distribution. Then we evaluate the output using the `model` function.

# %%
sampleSize_train = 20
X_train = myDistribution.getSample(sampleSize_train)
Y_train = model(X_train)

# %%
# The following figure presents the distribution of the vertical deviations Y on the training sample. We observe that the large deviations occur less often.
histo = ot.HistogramFactory().build(Y_train).drawPDF()
histo.setXTitle("Vertical deviation (cm)")
histo.setTitle("Distribution of the vertical deviation")
histo.setLegends([""])
view = otv.View(histo)

# %%
# Create the metamodel
# --------------------
#
# In order to create the GPR metamodel, we first select a constant trend with the :class:`~openturns.ConstantBasisFactory` class. Then we use a squared exponential covariance kernel.
# The :class:`~openturns.SquaredExponential` kernel has one amplitude coefficient and 4 scale coefficients.
# This is because this covariance kernel is anisotropic : each of the 4 input variables is associated with its own scale coefficient.
basis = ot.ConstantBasisFactory(cb.dim).build()
covarianceModel = ot.SquaredExponential(cb.dim)

# %%
# Finally, we use the :class:`~openturns.GaussianProcessFitter` and :class:`~openturns.GaussianProcessRegression` classes to create the GPR metamodel.
# It requires a training sample, a covariance kernel and a trend basis as input arguments.
fitter_algo = ot.GaussianProcessFitter(X_train, Y_train, covarianceModel, basis)

# %%
# The method :meth:`~openturns.GaussianProcessFitter.run`  of
# the class :class:`~openturns.GaussianProcessFitter` optimizes the Gaussian process
# hyperparameters and the method :meth:`~openturns.GaussianProcessRegression.run` of the class
# :class:`~openturns.GaussianProcessRegression` conditions the
# Gaussian process to the data set.
#
# We can then print the constant trend of the metamodel, estimated using the least squares method.
fitter_algo.run()
fitter_result = fitter_algo.getResult()
gpr_algo = ot.GaussianProcessRegression(fitter_result)
gpr_algo.run()
gpr_result = gpr_algo.getResult()
gprMetamodel = gpr_result.getMetaModel()

# %%
# The method :meth:`~openturns.GaussianProcessRegressionResult.getTrendCoefficients` of
# the class
# :class:`~openturns.GaussianProcessRegressionResult` returns the coefficients of the trend.
print(gpr_result.getTrendCoefficients())

# %%
# We can also print the hyperparameters of the covariance model, which have been estimated by maximizing the likelihood.
gpr_result.getCovarianceModel()

# %%
# Validate the metamodel
# ----------------------
#
# We finally want to validate the GPR metamodel. This is why we generate a validation sample with size 100 and we evaluate the output of the model on this sample.
sampleSize_test = 100
X_test = myDistribution.getSample(sampleSize_test)
Y_test = model(X_test)

# %%
# The class :class:`~openturns.MetaModelValidation`  makes the validation easy. To create it, we use the validation samples and the metamodel.
val = ot.MetaModelValidation(Y_test, gprMetamodel(X_test))

# %%
# The method :meth:`~openturns.MetaModelValidation.computeR2Score` computes the R2 score.
R2 = val.computeR2Score()[0]
print(R2)

# %%
# The residuals are the difference between the model and the metamodel.
r = val.getResidualSample()
graph = ot.HistogramFactory().build(r).drawPDF()
graph.setXTitle("Residuals (cm)")
graph.setTitle("Distribution of the residuals")
graph.setLegends([""])
view = otv.View(graph)

# %%
# We observe that the negative residuals occur with nearly the same frequency of the positive residuals: this is a first sign of good quality.
# The method :meth:`~openturns.MetaModelValidation.drawValidation` compares the observed outputs and the metamodel outputs.
graph = val.drawValidation()
graph.setTitle("R2 = %.2f%%" % (100 * R2))
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



"""
Export a field to VTK
=====================
"""

# %%
# The objective here is to create a field and export it as a VTK file.
# A field is the agregation of a mesh :math:`\mathcal{M}` of a domain :math:`\mathcal{D} \in \mathbb{R}^n`
# and a sample of values in :math:`\mathbb{R}^d` associated to each vertex of the mesh.
#
# We note :math:`(\vect{t}_0, \dots, \vect{t}_{N-1})` the vertices of :math:`\mathcal{M}`
# and :math:`(\vect{x}_0, \dots, \vect{x}_{N-1})` the associated values in :math:`\mathbb{R}^d`.
#
# A field is stored in the :class:`~openturns.Field` object that stores the mesh and the values at each vertex of the mesh.
# It can be built from a mesh and values or as a realization of a stochastic process.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# First, we define a regular 2-d mesh
discretization = [10, 5]
mesher = ot.IntervalMesher(discretization)
lowerBound = [0.0, 0.0]
upperBound = [2.0, 1.0]
interval = ot.Interval(lowerBound, upperBound)
mesh = mesher.build(interval)
graph = mesh.draw()
graph.setTitle("Regular 2-d mesh")
view = otv.View(graph)

# %%
# We now create a field from a mesh and some values
values = ot.Normal([0.0] * 2, [1.0] * 2, ot.CorrelationMatrix(2)).getSample(
    len(mesh.getVertices())
)
for i in range(len(values)):
    x = values[i]
    values[i] = 0.05 * x / x.norm()
field = ot.Field(mesh, values)

# %%
# We can export the `field` to a VTK files. It can be
# read later with an external program such as Paraview.
field.exportToVTKFile("field.vtk")

# %%
# Display all figures
otv.View.ShowAll()



"""
Apply a transform or inverse transform on your polynomial chaos
===============================================================
"""

# %%
#
# Introduction
# ------------
#
# In this example, we present the transformation involved in the creation of a polynomial chaos.
# Indeed, the polynomial chaos expansion is never directly applied to the input random variable of a model :math:`g`.
# Instead, the expansion is expressed based on the associated standard random variable.
# In this example,  we show how to transform a Log-Normal random variable into its standardized variable with the class :class:`~openturns.DistributionTransformation`.

# %%
# Probabilistic transform
# -----------------------
#
# Let :math:`T` be the probabilistic transform which maps the physical random variable :math:`X` to the standardized random variable :math:`\xi`:
#
# .. math::
#    \xi = T(X).
#
#
# Let :math:`F_X` be the Cumulative Distribution Function (CDF) associated with
# the random variable :math:`X` and let :math:`F_\xi` be the CDF associated
# with the random variable :math:`\xi`. Therefore, :math:`F_\xi` is the CDF
# associated with the orthogonal polynomials involved in the polynomial chaos expansion.
# In this case, the transform is:
#
# .. math::
#    \xi = F_\xi^{-1} \left( F_X(X) \right),
#
#
# for any :math:`X \in \mathbb{R}`.

# %%
# Example
# -------
#
# We want to use the Hermite orthogonal polynomials to expand a Log-Normal random variable:
#
# * let :math:`X` follow the Log-Normal distribution with the following parameters: LogNormal(:math:`\mu=3 \times 10^4`, :math:`\sigma=9\times 10^3`),
# * let :math:`Z=\xi` follow the Normal distribution with zero mean and unit standard deviation (the letter :math:`Z` is often used for standard Normal random variables).
#
# Let :math:`F_{LN}` be the CDF of the Log-Normal distribution associated with :math:`X` and let :math:`\Phi` be the CDF of the standard Normal distribution.
#
# Therefore,
#
# .. math::
#    Z = T(X) = \Phi^{-1}(F_{LN}(X)),
#
#
# for any :math:`X \in \mathbb{R}`.

# %%
import openturns as ot

# %%
# In the first step, we define the Log-Normal distribution.
# Its parameters - mean and standard deviation - have been selected so that there is no ambiguity with the standard Normal distribution.
# This parametrization can be used thanks to the :class:`~openturns.ParametrizedDistribution` class.

# %%
Xparam = ot.LogNormalMuSigma(3.0e4, 9.0e3, 15.0e3)  # in N
X = ot.ParametrizedDistribution(Xparam)

# %%
# Then we generate a 5-point sample.

# %%
sampleX = X.getSample(5)
sampleX

# %%
# In the second step, we define the random variable :math:`Z` with standard Normal distribution.

# %%
Z = ot.Normal()

# %%
# We also generate a sample from it.

# %%
sampleZ = Z.getSample(5)
sampleZ

# %%
# In the third step, we create the transform :math:`T` which maps :math:`X` to :math:`Z`.

# %%
T = ot.DistributionTransformation(X, Z)

# %%
# We apply this transform on the sample generated from :math:`X`.

# %%
T(sampleX)

# %%
# The inverse transform maps :math:`Z` into :math:`X`.

# %%
Tinverse = T.inverse()

# %%
Tinverse(sampleZ)

# %%
# Conclusion
# ----------
#
# The :class:`~openturns.DistributionTransformation` class is rarely used directly because the polynomial chaos classes perform the transformation automatically.
# However, understanding how this transform is done clarifies why the coefficients of the chaos expansion cannot be related to the input random variable :math:`X`
# the expansion is based on the standard variables :math:`\xi`.
# Hence, the absolute values of the corresponding coefficients have no straightforward interpretation,
# even though squaring them yields the part of the global variance associated with each coefficient.



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Tmin , time step and step number for TimeGrid
Tmin = 0.0
deltaT = 1.0
N = 11
timeGrid = ot.RegularGrid(Tmin, deltaT, N)

# Fixing the distribution for the RandomWalk
dist = ot.Distribution(ot.Uniform())
print("dist = ", dist)
origin = ot.Point(dist.getDimension())

# Building a random walk
process = ot.RandomWalk(origin, dist)

# print the process
process.setTimeGrid(timeGrid)
print("process = ", process)

# Get a realization of the process
timeSerie = process.getRealization()
print("timeSerie = ", timeSerie)

# Some steps further
stepNumber = 4
print("One future=", process.getFuture(stepNumber))
size = 3
print("Some futures=", process.getFuture(stepNumber, size))



#! /usr/bin/env python

import openturns as ot


class RandomExp:
    def __init__(self, size, dim):
        self.size = size
        self.dim = dim

    def generate(self):
        res = ot.Sample(self.size, self.dim)
        for i in range(self.size):
            for j in range(self.dim):
                res[i, j] = ot.RandomGenerator.Generate()
        return res


pyexp = RandomExp(10, 2)
experiment = ot.Experiment(pyexp)
print(experiment)
sample = experiment.generate()
print(sample)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

size = 10000
distribution = ot.GeneralizedExtremeValue(2.0, 1.5, -0.15)
sample = distribution.getSample(size)
# fake result
distributionParameter = ot.Normal(distribution.getParameter(), ot.CovarianceMatrix(3))
result = ot.DistributionFactoryResult(distribution, distributionParameter)
validation = ot.GeneralizedExtremeValueValidation(result, sample)
graph1 = validation.drawReturnLevel()
assert len(graph1.getDrawables()) == 4, "nb drawables"
graph2 = validation.drawPDF()
assert len(graph2.getDrawables()) == 2, "nb drawables"
grid = validation.drawDiagnosticPlot()
assert grid.getNbColumns() == 2, "nb cols"
assert grid.getNbRows() == 2, "nb rows"



import openturns as ot

u = 0.5
r = 2
sample = ot.Sample.BuildFromPoint([1, 2, -4, -5, 3, 4, -2, 5, -1, -3, 6, -2, -2, 7, 8])
part = ot.SamplePartition(sample)
peaks, clusters = part.getPeakOverThreshold(u, r)
print(peaks)
assert peaks.asPoint() == [2.0, 5.0, 6.0, 8.0]
indices = clusters.getIndicesCollection()
print(indices)
assert len(peaks) == len(indices)

graph = clusters.draw(u)



import openturns as ot
import openturns.viewer as otv

distribution_2d = ot.MarshallOlkinCopula(0.3, 0.7)
title = f"Marshall-Olkin copula CDF and sample: alpha = {distribution_2d.getAlpha()}, beta = {distribution_2d.getBeta()}"

graph = ot.Graph(title, "x1", "x2")
graph.add(ot.Cloud(distribution_2d.getSample(1000)))
graph.add(distribution_2d.drawCDF([0.0] * 2, [1.0] * 2, [71] * 2))
v = otv.View(graph)
v.ShowAll()



"""
Sequentially adding new points to a Gaussian Process metamodel
==============================================================
"""

# %%
# In this example, we show how to sequentially add new points to a  Gaussian Process fitter (Kriging) in order to improve the predictivity of the metamodel.
# In order to create simple graphics, we consider a 1-d function.

# %%
# Create the function and the design of experiments
# -------------------------------------------------

# %%
import openturns as ot
import openturns.viewer as otv
import numpy as np

# %%
sampleSize = 4
dimension = 1

# %%
# Define the function.

# %%
g = ot.SymbolicFunction(["x"], ["0.5*x^2 + sin(2.5*x)"])

# %%
# Create the design of experiments.

# %%
xMin = -0.9
xMax = 1.9
X_distr = ot.Uniform(xMin, xMax)
X = ot.LHSExperiment(X_distr, sampleSize, False, False).generate()
Y = g(X)

# %%
graph = g.draw(xMin, xMax)
data = ot.Cloud(X, Y)
data.setColor("red")
graph.add(data)
view = otv.View(graph)


# %%
# Create the algorithms
# ---------------------


# %%
def createMyBasicGPfitter(X, Y):
    """
    Create a Gaussian Process model from a pair of X and Y samples.
    We use a 3/2 Matérn covariance model and a constant trend.
    """
    basis = ot.ConstantBasisFactory(dimension).build()
    covarianceModel = ot.MaternModel([1.0], 1.5)
    fitter = ot.GaussianProcessFitter(X, Y, covarianceModel, basis)
    fitter.run()
    algo = ot.GaussianProcessRegression(fitter.getResult())
    algo.run()
    gprResult = algo.getResult()
    return gprResult


# %%
def linearSample(xmin, xmax, npoints):
    """Returns a sample created from a regular grid
    from xmin to xmax with npoints points."""
    step = (xmax - xmin) / (npoints - 1)
    rg = ot.RegularGrid(xmin, step, npoints)
    vertices = rg.getVertices()
    return vertices


# %%
# The following `sqrt` function will be used later to compute the standard deviation from the variance.

# %%
sqrt = ot.SymbolicFunction(["x"], ["sqrt(x)"])


# %%
def plotMyBasicGPfitter(gprResult, xMin, xMax, X, Y, level=0.95):
    """
    Given a metamodel result, plot the data, the GP fitter metamodel
    and a confidence interval.
    """
    samplesize = X.getSize()
    meta = gprResult.getMetaModel()
    graphKriging = meta.draw(xMin, xMax)
    graphKriging.setLegends(["Gaussian Process Regression"])
    # Create a grid of points and evaluate the function and the metamodel
    nbpoints = 50
    xGrid = linearSample(xMin, xMax, nbpoints)
    yFunction = g(xGrid)
    yKrig = meta(xGrid)
    # Compute the conditional covariance
    gpcc = ot.GaussianProcessConditionalCovariance(gprResult)
    epsilon = ot.Sample(nbpoints, [1.0e-8])
    conditionalVariance = gpcc.getConditionalMarginalVariance(xGrid) + epsilon
    conditionalSigma = sqrt(conditionalVariance)
    # Compute the quantile of the Normal distribution
    alpha = 1 - (1 - level) / 2
    quantileAlpha = ot.DistFunc.qNormal(alpha)
    # Graphics of the bounds
    epsilon = 1.0e-8
    dataLower = [
        yKrig[i, 0] - quantileAlpha * conditionalSigma[i, 0] for i in range(nbpoints)
    ]
    dataUpper = [
        yKrig[i, 0] + quantileAlpha * conditionalSigma[i, 0] for i in range(nbpoints)
    ]
    # Compute the Polygon graphics
    boundsPoly = ot.Polygon.FillBetween(xGrid.asPoint(), dataLower, dataUpper)
    boundsPoly.setLegend("95% bounds")
    # Validate the metamodel
    metamodelPredictions = meta(xGrid)
    mmv = ot.MetaModelValidation(yFunction, metamodelPredictions)
    r2Score = mmv.computeR2Score()[0]
    # Plot the function
    graphFonction = ot.Curve(xGrid, yFunction)
    graphFonction.setLineStyle("dashed")
    graphFonction.setColor("magenta")
    graphFonction.setLineWidth(2)
    graphFonction.setLegend("Function")
    # Draw the X and Y observed
    cloudDOE = ot.Cloud(X, Y)
    cloudDOE.setPointStyle("circle")
    cloudDOE.setColor("red")
    cloudDOE.setLegend("Data")
    # Assemble the graphics
    graph = ot.Graph()
    graph.add(boundsPoly)
    graph.add(graphFonction)
    graph.add(cloudDOE)
    graph.add(graphKriging)
    graph.setLegendPosition("lower right")
    graph.setAxes(True)
    graph.setGrid(True)
    graph.setTitle("Size = %d, R2=%.2f%%" % (samplesize, 100 * r2Score))
    graph.setXTitle("X")
    graph.setYTitle("Y")
    return graph


# %%
# We start by creating the initial GP fitter metamodel on the 4 points in the design of experiments.

# %%
gprResult = createMyBasicGPfitter(X, Y)
graph = plotMyBasicGPfitter(gprResult, xMin, xMax, X, Y)
view = otv.View(graph)


# %%
# Sequentially add new points
# ---------------------------

# %%
# The following function is the building block of the algorithm. It returns a new point which maximizes the conditional variance.


# %%
def getNewPoint(xMin, xMax, gprResult):
    """
    Returns a new point to be added to the design of experiments.
    This point maximizes the conditional variance of the metamodel.
    """
    nbpoints = 50
    xGrid = linearSample(xMin, xMax, nbpoints)
    gpcc = ot.GaussianProcessConditionalCovariance(gprResult)
    conditionalVariance = gpcc.getConditionalMarginalVariance(xGrid)
    iMaxVar = int(np.argmax(conditionalVariance))
    xNew = xGrid[iMaxVar, 0]
    xNew = ot.Point([xNew])
    return xNew


# %%
# We first call `getNewPoint` to get a point to add to the design of experiments.


# %%
xNew = getNewPoint(xMin, xMax, gprResult)
xNew

# %%
# Then we evaluate the function on the new point and add it to the training design of experiments.

# %%
yNew = g(xNew)
X.add(xNew)
Y.add(yNew)

# %%
# We now plot the updated GP fitter.

# %%
# sphinx_gallery_thumbnail_number = 3
gprResult = createMyBasicGPfitter(X, Y)
graph = plotMyBasicGPfitter(gprResult, xMin, xMax, X, Y)
graph.setTitle("GP fitter #0")
view = otv.View(graph)

# %%
# The algorithm added a point to the right bound of the domain.

# %%
for krigingStep in range(5):
    xNew = getNewPoint(xMin, xMax, gprResult)
    yNew = g(xNew)
    X.add(xNew)
    Y.add(yNew)
    gprResult = createMyBasicGPfitter(X, Y)
    graph = plotMyBasicGPfitter(gprResult, xMin, xMax, X, Y)
    graph.setTitle("GP fitter #%d " % (krigingStep + 1) + graph.getTitle())
    otv.View(graph)

# %%
# We observe that the second added point is the left bound of the domain.
# The remaining points were added strictly inside the domain where the accuracy was drastically improved.
#
# With only 10 points, the metamodel accuracy is already very good with a :math:`Q^2` which is equal to 99.9%.

# %%
# Conclusion
# ----------
#
# The current example presents the naive implementation on the creation of a sequential design of experiments based on Gaussian Process metamodel.
# More practical algorithms are presented in [ginsbourger2018]_.

otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Instantiate one distribution object
R = ot.CorrelationMatrix(3)
R[0, 1] = 0.5
R[0, 2] = 0.25
collection = [ot.FrankCopula(3.0), ot.NormalCopula(R), ot.ClaytonCopula(2.0)]
copula = ot.BlockIndependentCopula(collection)

print("Copula ", repr(copula))
print("Copula ", copula)

# Is this copula elliptical ?
print("Elliptical distribution= ", copula.isElliptical())

# Is this copula continuous ?
print("Continuous = ", copula.isContinuous())

# Is this copula elliptical ?
print("Elliptical = ", copula.hasEllipticalCopula())

# Is this copula independent ?
print("Independent = ", copula.hasIndependentCopula())

# Test for realization of copula
oneRealization = copula.getRealization()
print("oneRealization=", repr(oneRealization))

# Test for sampling
size = 10000
oneSample = copula.getSample(size)
print("oneSample first=", repr(oneSample[0]), " last=", repr(oneSample[size - 1]))
print("mean=", repr(oneSample.computeMean()))
print("covariance=", repr(oneSample.computeCovariance()))

# Define a point
point = ot.Point(copula.getDimension(), 0.6)
print("Point= ", repr(point))

# Show PDF and CDF of point
# Scalar eps(1e-5)
DDF = copula.computeDDF(point)
print("ddf     =", repr(DDF))
PDF = copula.computePDF(point)
print("pdf     =%.6f" % PDF)
CDF = copula.computeCDF(point)
print("cdf=%.6f" % CDF)
#    Point PDFgr = copula.computePDFGradient( point )
#    print "pdf gradient     =", PDFgr
#    Point CDFgr = copula.computeCDFGradient( point )
#    print "cdf gradient     =", CDFgr
quantile = copula.computeQuantile(0.95)
print("quantile=", repr(quantile))
print("cdf(quantile)=%.6f" % copula.computeCDF(quantile))
# Get 95% survival function
inverseSurvival = ot.Point(copula.computeInverseSurvivalFunction(0.95))
print("InverseSurvival=", repr(inverseSurvival))
print(
    "Survival(inverseSurvival)=%.6f" % copula.computeSurvivalFunction(inverseSurvival)
)
print("entropy=%.6f" % copula.computeEntropy())
# Confidence regions
if copula.getDimension() <= 2:
    threshold = ot.Point()
    print(
        "Minimum volume interval=", copula.computeMinimumVolumeInterval(0.95, threshold)
    )
    print("threshold=", threshold)
    beta = ot.Point()
    levelSet = copula.computeMinimumVolumeLevelSet(0.95, beta)
    print("Minimum volume level set=", levelSet)
    print("beta=", beta)
    print(
        "Bilateral confidence interval=",
        copula.computeBilateralConfidenceInterval(0.95, beta),
    )
    print("beta=", beta)
    print(
        "Unilateral confidence interval (lower tail)=",
        copula.computeUnilateralConfidenceInterval(0.95, False, beta),
    )
    print("beta=", beta)
    print(
        "Unilateral confidence interval (upper tail)=",
        copula.computeUnilateralConfidenceInterval(0.95, True, beta),
    )
    print("beta=", beta)

mean = copula.getMean()
print("mean=", repr(mean))
covariance = copula.getCovariance()
print("covariance=", repr(covariance))
parameters = copula.getParametersCollection()
print("parameters=", repr(parameters))
# Covariance and correlation
precision = ot.PlatformInfo.GetNumericalPrecision()
ot.PlatformInfo.SetNumericalPrecision(4)
covariance = copula.getCovariance()
print("covariance=", covariance)
correlation = copula.getCorrelation()
print("correlation=", correlation)
spearman = copula.getSpearmanCorrelation()
print("spearman=", spearman)
kendall = copula.getKendallTau()
print("kendall=", kendall)
ot.PlatformInfo.SetNumericalPrecision(precision)
dim = copula.getDimension()
x = 0.6
y = [0.2] * (dim - 1)
print("conditional PDF=%.6f" % copula.computeConditionalPDF(x, y))
print("conditional CDF=%.6f" % copula.computeConditionalCDF(x, y))
print("conditional quantile=%.6f" % copula.computeConditionalQuantile(x, y))
pt = ot.Point([0.1 * i + 0.05 for i in range(dim)])
print("sequential conditional PDF=", copula.computeSequentialConditionalPDF(point))
resCDF = copula.computeSequentialConditionalCDF(pt)
print("sequential conditional CDF(", pt, ")=", resCDF)
print(
    "sequential conditional quantile(",
    resCDF,
    ")=",
    copula.computeSequentialConditionalQuantile(resCDF),
)

# Specific to this copula

# Extract a 5-D marginal
dim = 5
point = ot.Point(dim, 0.25)
indices = ot.Indices(dim, 0)
indices[0] = 1
indices[1] = 2
indices[2] = 3
indices[3] = 5
indices[4] = 6
print("indices=", repr(indices))
margins = copula.getMarginal(indices)
print("margins=", repr(margins))
print("margins PDF=%.6f" % margins.computePDF(point))
print("margins CDF=%.6f" % margins.computeCDF(point))
quantile = margins.computeQuantile(0.95)
print("margins quantile=", repr(quantile))
print("margins CDF(quantile)=%.6f" % margins.computeCDF(quantile))
print("margins realization=", repr(margins.getRealization()))
# Tests the isoprobabilistic transformation
# General case with normal standard distribution
print(
    "isoprobabilistic transformation (general normal)=",
    copula.getIsoProbabilisticTransformation(),
)
# General case with non-normal standard distribution
collection[0] = ot.SklarCopula(
    ot.Student(3.0, ot.Point(2, 1.0), ot.Point(2, 3.0), ot.CorrelationMatrix(2))
)
copula = ot.BlockIndependentCopula(collection)
print(
    "isoprobabilistic transformation (general non-normal)=",
    copula.getIsoProbabilisticTransformation(),
)
dim = copula.getDimension()
x = 0.6
y = [0.2] * (dim - 1)
print("conditional PDF=%.6f" % copula.computeConditionalPDF(x, y))
print("conditional CDF=%.6f" % copula.computeConditionalCDF(x, y))
print("conditional quantile=%.6f" % copula.computeConditionalQuantile(x, y))
pt = ot.Point([0.1 * i + 0.05 for i in range(dim)])
print("sequential conditional PDF=", copula.computeSequentialConditionalPDF(pt))
resCDF = copula.computeSequentialConditionalCDF(pt)
print("sequential conditional CDF(", pt, ")=", resCDF)
print(
    "sequential conditional quantile(",
    resCDF,
    ")=",
    copula.computeSequentialConditionalQuantile(resCDF),
)
# Special case, independent copula
collection[0] = ot.SklarCopula(ot.Normal(2))
collection[1] = ot.IndependentCopula(2)
collection[2] = ot.NormalCopula(ot.CorrelationMatrix(2))
copula = ot.BlockIndependentCopula(collection)
print(
    "isoprobabilistic transformation (independent)=",
    copula.getIsoProbabilisticTransformation(),
)
dim = copula.getDimension()
x = 0.6
y = [0.2] * (dim - 1)
print("conditional PDF=%.6f" % copula.computeConditionalPDF(x, y))
print("conditional CDF=%.6f" % copula.computeConditionalCDF(x, y))
print("conditional quantile=%.6f" % copula.computeConditionalQuantile(x, y))
pt = ot.Point([0.1 * i + 0.05 for i in range(dim)])
print("sequential conditional PDF=", copula.computeSequentialConditionalPDF(pt))
resCDF = copula.computeSequentialConditionalCDF(pt)
print("sequential conditional CDF(", pt, ")=", resCDF)
print(
    "sequential conditional quantile(",
    resCDF,
    ")=",
    copula.computeSequentialConditionalQuantile(resCDF),
)
# Special case, single contributor
collection = [
    ot.SklarCopula(
        ot.Student(3.0, ot.Point(2, 1.0), ot.Point(2, 3.0), ot.CorrelationMatrix(2))
    )
]
copula = ot.BlockIndependentCopula(collection)
print(
    "isoprobabilistic transformation (single contributor)=",
    copula.getIsoProbabilisticTransformation(),
)
dim = copula.getDimension()
x = 0.6
y = [0.2] * (dim - 1)
print("conditional PDF=%.6f" % copula.computeConditionalPDF(x, y))
print("conditional CDF=%.6f" % copula.computeConditionalCDF(x, y))
print("conditional quantile=%.6f" % copula.computeConditionalQuantile(x, y))
pt = ot.Point([0.1 * i + 0.05 for i in range(dim)])
print("sequential conditional PDF=", copula.computeSequentialConditionalPDF(pt))
resCDF = copula.computeSequentialConditionalCDF(pt)
print("sequential conditional CDF(", pt, ")=", resCDF)
print(
    "sequential conditional quantile(",
    resCDF,
    ")=",
    copula.computeSequentialConditionalQuantile(resCDF),
)

# test BlockIndependentCopula.getMarginal in reverse
copula = ot.BlockIndependentCopula([ot.IndependentCopula(2), ot.NormalCopula(2)])
print(copula.getMarginal([3, 2, 1, 0]))



#! /usr/bin/env python

import openturns as ot
import math as m
import dill

# ensures python code is included
dill.settings["recurse"] = True

ot.TESTPREAMBLE()


def g(X):
    a, b = X
    Y = [a * m.sin(b)]
    return Y


f = ot.PythonFunction(2, 1, g)
x = [4, 5]
print(f(x))

# save
study = ot.Study()
study.setStorageManager(ot.XMLStorageManager("pyf.xml"))
study.add("f", f)
study.save()



"""
Specify a simulation algorithm
==============================
"""

# %%
# In this example we are going to parameterize a simulation algorithm:
#
#  - parameters linked to the number of points generated
#  - the precision of the probability estimator
#  - the sample storage strategy
#  - using callbacks to monitor progress and stopping criteria.
#

# %%
import openturns as ot


# %%
# Create the joint distribution of the parameters.

# %%
distribution_R = ot.LogNormalMuSigma(300.0, 30.0, 0.0).getDistribution()
distribution_F = ot.Normal(75e3, 5e3)
marginals = [distribution_R, distribution_F]
distribution = ot.JointDistribution(marginals)

# %%
# Create the model.

# %%
model = ot.SymbolicFunction(["R", "F"], ["R-F/(pi_*100.0)"])

# %%
# Create the event whose probability we want to estimate.

# %%
vect = ot.RandomVector(distribution)
G = ot.CompositeRandomVector(model, vect)
event = ot.ThresholdEvent(G, ot.Less(), 0.0)

# %%
# Create a Monte Carlo algorithm.

# %%
experiment = ot.MonteCarloExperiment()
algo = ot.ProbabilitySimulationAlgorithm(event, experiment)

# %%
# Criteria 1: Define the Maximum Coefficient of variation of the probability estimator.

# %%
algo.setMaximumCoefficientOfVariation(0.05)

# %%
# Criteria 2: Define the number of iterations of the simulation.

# %%
algo.setMaximumOuterSampling(int(1e4))

# %%
# The block size parameter represents the number of samples evaluated per iteration, useful for parallelization.

# %%
algo.setBlockSize(2)

# %%
# HistoryStrategy to store the values of the probability used to draw the convergence graph.

# %%
# Null strategy
algo.setConvergenceStrategy(ot.Null())

# Full strategy
algo.setConvergenceStrategy(ot.Full())

# Compact strategy: N points
N = 1000
algo.setConvergenceStrategy(ot.Compact(N))


# %%
# Use a callback to display the progress every 10%.


# %%
def progress(p):
    if p >= progress.t:
        progress.t += 10.0
        print("progress=", p, "%")
    return False


progress.t = 10.0
algo.setProgressCallback(progress)


# %%
# Use a callback to stop the simulation.


# %%
def stop():
    # here we never stop, but we could
    return False


algo.setStopCallback(stop)

# %%
algo.run()

# %%
# Retrieve results.

# %%
result = algo.getResult()
probability = result.getProbabilityEstimate()
print("Pf=", probability)



#! /usr/bin/env python

import openturns as ot

ot.Log.Show(ot.Log.ALL)
distribution = ot.Frechet(2.5, 2.0, -1.5)
size = 10000
sample = distribution.getSample(size)
factory = ot.FrechetFactory()
print("Distribution                      =", repr(distribution))
result = factory.buildEstimator(sample)
estimatedDistribution = result.getDistribution()
print("Estimated distribution            =", repr(estimatedDistribution))
parameterDistribution = result.getParameterDistribution()
print("Parameter distribution            =", parameterDistribution)
defaultDistribution = factory.build()
print("Default distribution              =", defaultDistribution)
fromParameterDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters      =", fromParameterDistribution)
typedEstimatedDistribution = factory.buildAsFrechet(sample)
print("Typed estimated distribution      =", typedEstimatedDistribution)
defaultTypedDistribution = factory.buildAsFrechet()
print("Default typed distribution        =", defaultTypedDistribution)
typedFromParameterDistribution = factory.buildAsFrechet(distribution.getParameter())
print("Typed distribution from parameters=", typedFromParameterDistribution)

# More involved test: the sample distribution does not fit the factory

# The distributions used :
myFrechet = ot.Frechet(1.0, 1.0, 0.0)
myGumbel = ot.Gumbel(1.0, 3.0)
# We build our mixture sample of size 2*1000=2000.
mixtureSample = ot.Sample()
sampleFrechet = myFrechet.getSample(1000)
sampleGumbel = myGumbel.getSample(1000)
mixtureSample.add(sampleFrechet)
mixtureSample.add(sampleGumbel)
# Build on the mixture sample
typedEstimatedFromMixtureSample = factory.buildAsFrechet(mixtureSample)
print("Estimated dist from mixture sample=", typedEstimatedFromMixtureSample)



#! /usr/bin/env python

import openturns as ot
import openturns.experimental as otexp
import openturns.testing as ott

ot.TESTPREAMBLE()

inputDimension = 2
constant = [1, 2, 3]

f = otexp.ConstantFunction(inputDimension, constant)

print(f)
print(repr(f))

assert f.getInputDimension() == inputDimension
assert f.getOutputDimension() == len(constant)
assert f == f

x = [5, 6]
y = f(x)
print(y)
ott.assert_almost_equal(y, constant)

dx = f.gradient(x)
ott.assert_almost_equal(dx, ot.Matrix(inputDimension, len(constant)))

dx2 = f.hessian(x)
ott.assert_almost_equal(dx2, ot.SymmetricTensor(inputDimension, len(constant)))

f02 = f.getMarginal([0, 2])
print(f02)
y02 = f02(x)
ott.assert_almost_equal(y02, [1, 3])



"""
Deterministic design of experiments
===================================
"""

# %%
# In this example we present the available deterministic design of experiments.
#
# Four types of deterministic design of experiments are available:
#
# - `Axial`
# - `Factorial`
# - `Composite`
# - `Box`
#
# Each type of deterministic design is discretized differently according to a number of levels.
#
# Functionally speaking, a design is a `Sample` that lies within the unit cube :math:`(0,1)^d` and can be scaled and moved to cover the desired box.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# We will use the following function to plot bi-dimensional samples.


def drawBidimensionalSample(sample, title):
    n = sample.getSize()
    graph = ot.Graph("%s, size=%d" % (title, n), "X1", "X2")
    cloud = ot.Cloud(sample)
    graph.add(cloud)
    return graph


# %%
# Axial design
# ------------

# %%
levels = [1.0, 1.5, 3.0]
experiment = ot.Axial(2, levels)
sample = experiment.generate()
graph = drawBidimensionalSample(sample, "Axial")
view = otv.View(graph)

# %%
# Scale and to get desired location.

# %%
sample *= 2.0
sample += [5.0, 8.0]
graph = drawBidimensionalSample(sample, "Axial")
view = otv.View(graph)

# %%
# Factorial design
# ----------------
#

# %%
experiment = ot.Factorial(2, levels)
sample = experiment.generate()
sample *= 2.0
sample += [5.0, 8.0]
graph = drawBidimensionalSample(sample, "Factorial")
view = otv.View(graph)

# %%
# Composite design
# ----------------

# %%
experiment = ot.Composite(2, levels)
sample = experiment.generate()
sample *= 2.0
sample += [5.0, 8.0]
graph = drawBidimensionalSample(sample, "Composite")
view = otv.View(graph)

# %%
# Grid design
# -----------
#

# %%
levels = [3, 4]
experiment = ot.Box(levels)
sample = experiment.generate()
sample *= 2.0
sample += [5.0, 8.0]
graph = drawBidimensionalSample(sample, "Box")
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()
